from typing import Union

import cv2
import numpy as np

from byotrack import OpticalFlow


class OpenCVOpticalFlow(OpticalFlow):
    """Wraps Open-CV optical flow implementations

    Currently ByoTrack only supports dense optical flow which computes
    the displacement of every pixel of the reference image.

    Usage:

    .. code-block:: python

        import cv2
        from byotrack.implementation.optical_flow.opencv import OpenCVOpticalFlow

        # See the documentation of Open-CV for each algorithm to correctly set the parameters
        parameters = {}

        # Farneback
        optflow = OpenCVOpticalFlow(cv2.FarnebackOpticalFlow.create(**parameters))

        # TVL1 => It requires to install opencv-contrib-python (and not just opencv-python)
        optflow = OpenCVOpticalFlow(cv2.optflow.DualTVL1OpticalFlow.create(**parameters))

        # Other methods from opencv-contrib could probably be used (RLOF, DIS, ...) but have not been tested yet

    Attributes:
        optical_flow (cv2.DenseOpticalFlow): Any implementation of the cv2 DenseOpticalFlow class

    """

    def __init__(
        self,
        optical_flow: cv2.DenseOpticalFlow,
        downscale: Union[float, np.ndarray] = 2,
        blur: Union[None, float, np.ndarray] = None,
    ):
        super().__init__(downscale, blur)
        self.optical_flow = optical_flow

    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        # Average channels and go to uint8 (multi channel is not supported)
        reference = np.round(reference.mean(axis=-1) * 255).astype(np.uint8)
        moving = np.round(moving.mean(axis=-1) * 255).astype(np.uint8)
        flow_map: np.ndarray = self.optical_flow.calc(reference, moving, None)  # type: ignore

        # Convert the flow map into (i, j) coordinates and (2, H', W')
        # In open-cv flow maps are (x, y) and of size (H', W', 2)
        flow_map = flow_map.transpose(2, 0, 1)
        flow_map = flow_map[::-1]
        return np.ascontiguousarray(flow_map)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Iterable, List, Sequence, Union

import numpy as np
import tqdm.auto as tqdm

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class Tracker(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base abstract tracker class

    A tracker can be run on a whole video and produces tracks

    """

    @abstractmethod
    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        """Run the tracker on a whole video

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Sequence of T frames (array).
                Each array is expected to have a shape (H, W, C)

        Returns:
            Collection[byotrack.Track]: Tracks of particles

        """


class MultiStepTracker(Tracker):  # pylint: disable=too-few-public-methods
    """Multi step tracker: split the tracking task into Detect / Link / Refine

    Attributes:
        detector (byotrack.Detector): Performs the detection on the video
        linker (byotrack.Linker): Links detections through time
        refiners (Sequence[byotrack.Refiner]): Optional refinement steps
            Empty by default

    """

    def __init__(
        self, detector: byotrack.Detector, linker: byotrack.Linker, refiners: Iterable[byotrack.Refiner] = ()
    ) -> None:
        super().__init__()

        self.detector = detector
        self.linker = linker
        self.refiners = refiners

    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        detections = self.detector.run(video)
        tracks = self.linker.run(video, detections)
        for refiner in self.refiners:
            tracks = refiner.run(video, tracks)

        return tracks


class BatchMultiStepTracker(MultiStepTracker):  # pylint: disable=too-few-public-methods
    """Online Tracking with detect/link/refine framework

    It only works with BatchDetector and OnlineLinker: it reduces RAM usage and computations
    by loading frames only by batch. Each batch is processed by the detector followed
    by the linker. Once a batch is processed, a new one is loaded. The detections are never fully stored,
    neither the video.

    Tracks refining is done offline.

    Attributes:
        detector (byotrack.BatchDetector): Performs the detection on the video by batch
            It defines the batch size to use
        linker (byotrack.OnlineLinker): Links detections one frame at time
        refiners (Sequence[byotrack.Refiner]): Optional refinement steps
            Empty by default

    """

    def __init__(
        self, detector: byotrack.BatchDetector, linker: byotrack.OnlineLinker, refiners: Iterable[byotrack.Refiner] = ()
    ) -> None:
        super().__init__(detector, linker, refiners)
        self.detector: byotrack.BatchDetector
        self.linker: byotrack.OnlineLinker

    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        reader = None
        if self.detector.add_true_frames and isinstance(video, byotrack.Video):
            reader = video.reader

        self.linker.reset()

        detections_sequence: Sequence[byotrack.Detections] = []
        batch: List[np.ndarray] = []
        frame_ids: List[int] = []

        detect_bar = tqdm.tqdm(desc=self.detector.progress_bar_description, total=len(video))
        link_bar = tqdm.tqdm(desc=self.linker.progress_bar_description, total=len(video))

        for i, frame in enumerate(video):
            batch.append(frame[None])
            frame_ids.append(reader.tell() if reader else i)

            if len(batch) >= self.detector.batch_size:
                detections_sequence = self.detector.detect(np.concatenate(batch, axis=0))
                detect_bar.update(self.detector.batch_size)

                for frame_id, frame, detections in zip(frame_ids, batch, detections_sequence):
                    detections.frame_id = frame_id
                    self.linker.update(frame[0], detections)
                    link_bar.update()

                link_bar.refresh()

                batch = []
                frame_ids = []

        if batch:
            detections_sequence = self.detector.detect(np.concatenate(batch, axis=0))
            detect_bar.update(len(batch))

            for frame_id, frame, detections in zip(frame_ids, batch, detections_sequence):
                detections.frame_id = frame_id
                self.linker.update(frame[0], detections)
                link_bar.update()

        detect_bar.close()
        link_bar.close()

        tracks = self.linker.collect()
        for refiner in self.refiners:
            tracks = refiner.run(video, tracks)

        return tracks



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Sequence, Union

import numpy as np
import tqdm.auto as tqdm

import byotrack  # pylint: disable=cyclic-import
from ..parameters import ParametrizedObjectMixin


class Detector(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base class for detections in videos

    Detect on each frame the objects of interest.

    Each detector can define a set of parameters (See `ParametrizedObjectMixin`)

    """

    @abstractmethod
    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Sequence[byotrack.Detections]:
        """Run the detector on a whole video

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Sequence of T frames (array).
                Each array is expected to have a shape (H, W, C)

        Returns:
            Sequence[byotrack.Detections]: Detections for each frame (ordered by frames)

        """


class BatchDetector(Detector):
    """Abstract detector that performs detection directly by batch

    Usually leads to faster implementation of the detection process
    when batch size is greater than 1

    Attributes:
        batch_size (int): Size of the frame batch
            Default: 20
        add_true_frames (bool): If the input is a Video, it will exploits the VideoReader knowledge
            to extract the true frame id for each detections.
            If False, it will register the index of the frame as the frame_id
            Default: False

    """

    progress_bar_description = "Detections"

    def __init__(self, batch_size=20, add_true_frames=False) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.add_true_frames = add_true_frames

    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> List[byotrack.Detections]:
        reader = None
        if self.add_true_frames and isinstance(video, byotrack.Video):
            reader = video.reader

        detections_sequence: List[byotrack.Detections] = []
        batch: List[np.ndarray] = []
        frame_ids: List[int] = []

        for i, frame in enumerate(tqdm.tqdm(video, self.progress_bar_description, miniters=self.batch_size)):
            batch.append(frame[None, ...])
            frame_ids.append(reader.tell() if reader else i)

            if len(batch) >= self.batch_size:
                detections_sequence.extend(self.detect(np.concatenate(batch, axis=0)))
                batch = []

        if batch:
            detections_sequence.extend(self.detect(np.concatenate(batch, axis=0)))

        # Set frame ids
        for frame_id, detections in zip(frame_ids, detections_sequence):
            detections.frame_id = frame_id

        return detections_sequence

    @abstractmethod
    def detect(self, batch: np.ndarray) -> Sequence[byotrack.Detections]:
        """Apply the detection on a batch of frames

        By default, the frame ids are set from 0 to n-1 with n the size of the batch.
        The aggregattion of batches and frame ids correction is automatically handled when called
        the `run` method.

        Args:
            batch (np.ndarray): Batch of video frames
                Shape: (B, H, W, C)

        Returns:
            Sequence[byotrack.Detections]: Detections for each given frame

        """



"""Provide some example data to test ByoTrack"""

import os
import urllib.request

import platformdirs

import byotrack


data_dir = platformdirs.user_data_dir("ByoTrack", roaming=True)
os.makedirs(data_dir, exist_ok=True)


def hydra_neurons() -> byotrack.Video:
    """2D time-lapse sequence of Hydra Vulgaris neurons (TdTomato).

    This data comes from: Hanson, A., Reme, R., Telerman, N., Yamamoto, W., Olivo-Marin, J. C.,
    Lagache, T., & Yuste, R. (2024). "Automatic monitoring of neural activity with single-cell resolution
    in behaving Hydra". Scientific Reports, 14(1), 5083.

    Please cite the paper if you use this data.

    It downloads the tdt_contrxn-1.avi file (https://datadryad.org/stash/dataset/doi:10.5061/dryad.h9w0vt4q3)
    into your user data folder if not already downloaded. The file is then read as a Video.

    It this fails, you can download the data manually and then open it using byotrack.Video(path).

    Returns:
        byotrack.Video: Video of Hydra Vulgaris neurons.
            Shape: (T=1000, H=848, W=1024, C=3)

    """
    file = os.path.join(data_dir, "tdt_contrxn-1.avi")
    if not os.path.exists(file):
        print("Downloading data from https://github.com/raphaelreme/byotrack/ ...")
        urllib.request.urlretrieve(
            "https://github.com/raphaelreme/byotrack/raw/main/example_data/tdt_contrxn-1.avi", file
        )

    try:
        return byotrack.Video(file)
    except:
        print("The downloaded data may be corrupted. You can manually download it from the ByoTrack github.")
        raise



from __future__ import annotations

import os
import re
from typing import Dict, Iterable, List, Sequence, Tuple, Union

import numba  # type: ignore
import numpy as np
import torch


def sorted_alphanumeric(data: Iterable[str]):
    """Sorts alphanumeriacally an iterable of strings

    "1" < "2" < "10" < "foo1" < "foo2" < "foo3"

    """

    def alphanum_key(key: str) -> List[Union[str, int]]:
        def convert(text: str) -> Union[str, int]:
            return int(text) if text.isdigit() else text.lower()

        return [convert(text) for text in re.split("([0-9]+)", key)]

    return sorted(data, key=alphanum_key)


def _check_segmentation(segmentation: torch.Tensor) -> None:
    """Check segmentation validity (type, shape, labels)"""
    assert len(segmentation.shape) == 2
    assert segmentation.dtype is torch.int32

    # Check that labels are in [0:N+1]
    assert torch.unique(segmentation).shape[0] == segmentation.max() + 1, "Labels are not consecutive"


def _check_bbox(bbox: torch.Tensor) -> None:
    """Check box validity (type, shape)"""
    assert len(bbox.shape) == 2
    assert bbox.shape[1] == 4, "Bbox should have 4 values (top, left, height, width)"
    assert bbox.dtype is torch.int32


def _check_position(position: torch.Tensor) -> None:
    """Check position validity (type, shape)"""
    assert len(position.shape) == 2
    assert position.shape[1] == 2, "Position should have 2 values (i, j)"
    assert position.dtype is torch.float32


def _check_confidence(confidence: torch.Tensor) -> None:
    """Check confidence validity (type, shape)"""
    assert len(confidence.shape) == 1
    assert confidence.dtype is torch.float32


@numba.njit
def _position_from_segmentation(segmentation: np.ndarray) -> np.ndarray:
    """Return the centre of each instance in the segmentation"""
    n = segmentation.max()
    positions = np.zeros((n, 2), dtype=np.float32)
    m_00 = np.zeros(n, dtype=np.uint)
    m_01 = np.zeros(n, dtype=np.uint)
    m_10 = np.zeros(n, dtype=np.uint)

    for i in range(segmentation.shape[0]):
        for j in range(segmentation.shape[1]):
            instance = segmentation[i, j] - 1
            if instance != -1:
                m_00[instance] += 1
                m_01[instance] += i
                m_10[instance] += j

    positions[:, 0] = m_01 / m_00
    positions[:, 1] = m_10 / m_00
    return positions


@numba.njit
def _bbox_from_segmentation(segmentation: np.ndarray) -> np.ndarray:
    n = segmentation.max()
    bbox = np.zeros((n, 4), dtype=np.int32)
    mini = np.ones((n, 2), dtype=np.int32) * np.inf
    maxi = np.zeros((n, 2), dtype=np.int32)

    for i in range(segmentation.shape[0]):
        for j in range(segmentation.shape[1]):
            instance = segmentation[i, j] - 1
            if instance != -1:
                mini[instance] = min(mini[instance][0], i), min(mini[instance][1], j)
                maxi[instance] = max(maxi[instance][0], i), max(maxi[instance][1], j)

    bbox[:, :2] = mini
    bbox[:, 2:] = maxi - mini + 1
    return bbox


@numba.njit
def _segmentation_from_bbox(bbox: np.ndarray, shape: Tuple[int, int]) -> np.ndarray:
    segmentation = np.zeros(shape, dtype=np.int32)
    for label, bbox_ in enumerate(bbox):
        segmentation[bbox_[0] : bbox_[0] + bbox_[2], bbox_[1] : bbox_[1] + bbox_[3]] = label + 1

    return segmentation


def _segmentation_from_position(position: torch.Tensor, shape: Tuple[int, int]) -> torch.Tensor:
    segmentation = torch.zeros(shape, dtype=torch.int32)
    segmentation[position.round().int().T.tolist()] = torch.arange(1, position.shape[0] + 1, dtype=torch.int32)
    return segmentation


def relabel_consecutive(segmentation: torch.Tensor) -> torch.Tensor:
    """Relabel a segmentation mask so that labels are consecutives

    For N instances, labels are 0 for background and then [1:N] for each instance.

    Args:
        segmentation (torch.Tensor): Segmentation mask
            Shape: (H, W), dtype: int32

    Returns:
        torch.Tensor: The same segmentation mask where labels are consecutive (from 0 to N)

    """
    labels: torch.Tensor = torch.unique(segmentation)
    max_label = labels.max().item()

    assert isinstance(max_label, int)

    if max_label + 1 == labels.shape[0]:
        return segmentation  # Nothing to do

    # Has to go to int64 has torch does not support indexing with int32 yet
    mapping = torch.zeros(max_label + 1, dtype=torch.int64)
    mapping[labels.to(torch.int64)] = torch.arange(labels.shape[0])

    return mapping[segmentation.to(torch.int64)].to(torch.int32)


class Detections:
    """Detections for a given frame

    Built from a data dict. The data has to contained one of "position",
    "bbox" or "segmentation" keys that respectively define the positions of instances center
    (i, j), the bounding boxes of instances (top, left, height, width) or the instance
    segmentation of the image (H, W).

    Positions are stored as floats (row first).
    Bounding boxes are stored as ints (row first), thus right = left + width - 1
    The labels of the segmentation mask have to be consecutive. You can make it consecutive
    using `relabel_consecutive`.

    Additional optional data is also expected like "confidence" or "shape" that respectively
    defines the confidence for each detection and the shape of the image (H, W).

    Any additional meta information on the detections can be also given.

    Defines position, bbox, segmentation, and confidence properties. Each of them are either from
    the data or extrapolated if missing (see _extrapolate_xxx).

    Attributes:
        data (Dict[str, toch.Tensor]): Detections data.
        length (int): Number of detections
        frame_id (int): Optional frame id in the original video (-1 if no video)
            In ByoTrack, detections linking do not rely on this frame_id, but rather
            on the position inside the detections_sequence. It should only be used
            for debugging/visualization.
            Default: -1
        shape: (Tuple[int, int]): Shape of the image (H, W). (Extrapolated if not given)
        position (torch.Tensor): Positions (i, j) of instances (center) inferred from the data
            Shape: (N, 2), dtype: float32
        bbox (torch.Tensor): Bounding boxes of instances inferred from the data
            Shape: (N, 4), dtype: int32
        segmentation (torch.Tensor): Segmentation inferred from the data
            Shape: (H, W), dtype: int32
        confidence (torch.Tensor): Confidence for each instance
            Shape: (N,), dtype: float32

    """

    def __init__(self, data: Dict[str, torch.Tensor], frame_id: int = -1) -> None:
        self.length = -1

        if "position" in data:
            _check_position(data["position"])
            self.length = data["position"].shape[0]

        if "bbox" in data:
            _check_bbox(data["bbox"])
            length = data["bbox"].shape[0]

            assert self.length in (-1, self.length)
            self.length = length

        if "segmentation" in data:
            _check_segmentation(data["segmentation"])
            length = int(data["segmentation"].max())

            assert self.length in (-1, self.length)
            self.length = length

        if "condidence" in data:
            _check_confidence(data["confidence"])
            assert data["confidence"].shape[0] == self.length

        self.data = data
        self._lazy_extrapolated_data: Dict[str, torch.Tensor] = {}

        self.shape = self._extrapolate_shape()
        self.frame_id = frame_id

    @property
    def position(self) -> torch.Tensor:
        position = self.data.get("position", self._lazy_extrapolated_data.get("position"))
        if position is None:
            position = self._extrapolate_position()
            self._lazy_extrapolated_data["position"] = position

        return position

    @property
    def bbox(self) -> torch.Tensor:
        bbox = self.data.get("bbox", self._lazy_extrapolated_data.get("bbox"))
        if bbox is None:
            bbox = self._extrapolate_bbox()
            self._lazy_extrapolated_data["bbox"] = bbox

        return bbox

    @property
    def segmentation(self) -> torch.Tensor:
        segmentation = self.data.get("segmentation", self._lazy_extrapolated_data.get("segmentation"))
        if segmentation is None:
            segmentation = self._extrapolate_segmentation()
            self._lazy_extrapolated_data["segmentation"] = segmentation

        return segmentation

    @property
    def confidence(self) -> torch.Tensor:
        confidence = self.data.get("confidence", self._lazy_extrapolated_data.get("confidence"))
        if confidence is None:
            confidence = self._extrapolate_confidence()
            self._lazy_extrapolated_data["confidence"] = confidence

        return confidence

    def _extrapolate_shape(self) -> Tuple[int, int]:
        """Extrapolate shape from data

        Etiher given in data, or from the segmentation shape, or from the shape needed to fit positions/bboxes.

        (Stored in `self.shape`, not in `self._lazy_extrapolated_data`)

        """
        if "shape" in self.data:
            return (int(self.data["shape"][0]), int(self.data["shape"][1]))

        if "segmentation" in self.data:
            return (self.data["segmentation"].shape[0], self.data["segmentation"].shape[1])

        if self.length == 0:
            return (1, 1)  # No data to extrapolate it.

        if "bbox" in self.data:
            maxi = self.data["bbox"][:, :2] + self.data["bbox"][:, 2:]
            shape = maxi.max(dim=0).values
            return (int(shape[0]), int(shape[1]))

        shape = self.data["position"].max(dim=0).values.ceil().int() + 1
        return (int(shape[0]), int(shape[1]))

    def _extrapolate_position(self) -> torch.Tensor:
        """Extrapolate position from data

        Average the segmentation for each label if exists or uses the bbox

        """
        if "segmentation" in self.data:
            return torch.tensor(_position_from_segmentation(self.data["segmentation"].numpy()))

        return self.data["bbox"][:, :2] + (self.data["bbox"][:, 2:] - 1) / 2

    def _extrapolate_bbox(self) -> torch.Tensor:
        """Extrapolate bbox from data

        Bounding box of the segmentation if exists, or return a bbox of shape 1 centered on the position

        """
        if "segmentation" in self.data:
            return torch.tensor(_bbox_from_segmentation(self.data["segmentation"].numpy()))

        bbox = torch.ones((self.length, 4), dtype=torch.int32)
        bbox[:, :2] = self.data["position"].round().int()
        return bbox

    def _extrapolate_segmentation(self) -> torch.Tensor:
        """Extrapolate segmentation from data

        Fill the bounding boxes if exists, or fill one pixel on each position

        Warning: unable to handle overlapping bboxes/positions

        """
        if "bbox" in self.data:
            return torch.tensor(_segmentation_from_bbox(self.data["bbox"].numpy(), self.shape))

        return _segmentation_from_position(self.data["position"], self.shape)

    def _extrapolate_confidence(self) -> torch.Tensor:
        """Extrapolate confidence"""
        return torch.ones(self.length)

    def __len__(self) -> int:
        return self.length

    def save(self, path: Union[str, os.PathLike]) -> None:
        """Save detections to a file using `torch.save`

        Args:
            path (str | os.PathLike): Output path

        """
        torch.save({"frame_id": self.frame_id, **self.data}, path)

    @staticmethod
    def load(path: Union[str, os.PathLike]) -> Detections:
        """Load a detections for a given frame using `torch.load`

        Args:
            path (str | os.PathLike): Input path

        """
        data = torch.load(path, map_location="cpu")
        assert isinstance(data, dict)
        frame_id = data.pop("frame_id")

        return Detections(data, frame_id)

    @staticmethod
    def save_multi_frames_detections(detections_sequence: Sequence[Detections], path: Union[str, os.PathLike]) -> None:
        """Save detections for a sequence of frames

        It will save the detections as::

            path/{frame_id}.pt
                 ...


        Args:
            detections_sequence (Sequence[Detections]): Detections for each frame
                Each detections should have a different frame_id
            path (str | os.PathLike): Output folder

        """
        os.makedirs(path)

        for i, detections in enumerate(detections_sequence):
            detections.save(os.path.join(path, f"{i}.pt"))

    @staticmethod
    def load_multi_frames_detections(path: Union[str, os.PathLike]) -> List[Detections]:
        """Load detections for a sequence of frames

        Expect the following file structure::

            path/{0}.pt
                 ...
                 {i}.pt
                 ...
                 {n}.pt

        Args:
            path (str | os.PathLike): Input folder

        Returns:
            List[Detections]: Detections for each frame (sorted by frame id)

        """
        files = os.listdir(path)

        detections_sequence: List[Detections] = []

        for i, file in enumerate(filter(lambda file: file[-3:] == ".pt", sorted_alphanumeric(files))):
            assert file == f"{i}.pt", f"The {i}th file is not '{i}.pt'"
            detections_sequence.append(Detections.load(os.path.join(path, file)))

        return detections_sequence



from __future__ import annotations

import os
from typing import Dict, List, Tuple, Union

import cv2  # type: ignore
import numpy as np
from PIL import Image  # type: ignore


class MetaVideoReader(type):
    """MetaClass for Video Readers

    Each VideoReader has to define a list of supported extensions.
    The last constructed VideoReader to claim an extension will be used to open the video.
    If no one has claimed an extension the default OpenCVVideoReader is used.
    """

    supported_extensions: List[str] = []
    extension_to_reader: Dict[str, MetaVideoReader] = {}

    def __init__(cls, cls_name: str, bases: tuple, attributes: dict) -> None:
        super().__init__(cls_name, bases, attributes)

        for extension in cls.supported_extensions:
            type(cls).extension_to_reader[extension] = cls


class VideoReader(metaclass=MetaVideoReader):
    """Unified video reader api

    Close to OpenCV API but few key differences:

    * There is always a frame loaded
    * Frame ids goes from 0 to length - 1
    * Read method is very different:

        * It retrieves the current frame then grabs the next (The other way around in opencv)
        * It returns therefore a ndarray and a bool rather than a bool and a ndarray
        * The boolean returned indicated if we can continue to read and not if the read operation has failed
    * Easy to check main attributes like:

        * frame_id
        * length
        * shape
        * fps if known (-1 otherwise)

    Return images in BGR by default like opencv as frames are mostly used with opencv afterwards for display.
    Can also return grayscale image (H, W, 1)

    Attributes:
        supported_extensions (List[str]): Static attribute used by `open` method to automatically choose
            which VideoReader to use.
        path (str | bytes | os.PathLike): Path of the current video
        released (bool): True when release has been called (close and release memory)
        fps (int): Frame rate (-1 if unknown)
        shape (Tuple[int, ...]): Spatial dimensions of frames (Height, Width[, Depth])
        channels (int): Number of channels
        length (int): Number of frames
        frame_id (int): Current frame id

    """

    supported_extensions: List[str] = []

    def __init__(self, path: Union[str, os.PathLike], **kwargs):  # pylint: disable=unused-argument
        """Constructor: Open the video file

        Args:
            path (str | os.PathLike): Path to the video to read
            kwargs: Any additional kwargs are given to the underlying video reader

        """
        self.path = path
        self.released = False
        self.fps = -1
        self.shape: Tuple[int, ...] = (0, 0)
        self.channels = 0
        self.length = 0
        self.frame_id = 0

    def release(self) -> None:
        """Close the file and free memory"""
        self.released = True

    def grab(self) -> bool:
        """Grab the next frame

        Can be faster than self.seek(self.frame_id + 1)

        Returns:
            bool: True if able to grab next frame

        """
        raise NotImplementedError()

    def retrieve(self) -> np.ndarray:
        """Retrieve the current frame

        Returns:
            np.ndarray: The current frame
                Shape: (H, W, 3) or (H, W, 1) (Grayscale)

        """
        raise NotImplementedError()

    def read(self) -> Tuple[np.ndarray, bool]:
        """Consume a frame. Is equivalent to retrieve + grab

        As in this implementation there is always a current frame. It reverses open cv implementation
        It first retrieves then grab next frame

        Returns:
            np.ndarray: The current frame
                Shape: (H, W, 3) or (H, W, 1) (Grayscale)
            bool: Whether there is a next frame to read

        """
        current_frame = self.retrieve()

        return current_frame, self.grab()

    def seek(self, frame_id: int) -> None:
        """Seek frame_id (will update the current frame)

        Valid frame ids from 0 to length - 1

        Raise:
            EOFError if seeking an invalid frame

        """
        raise NotImplementedError()

    def tell(self) -> int:
        """Returns self.frame_id

        Returns:
            int: current frame_id

        """
        return self.frame_id

    def _check_frame_id(self) -> None:
        """Check that the underlying data is at the right frame id"""

    @staticmethod
    def open(path: Union[str, os.PathLike], **kwargs) -> VideoReader:
        """Open a video file

        Use the extension to know which VideoReader to use

        Args:
            path (str | os.PathLike): File to open
            kwargs: Any additional args for the underlying video reader

        Returns:
            VideoReader

        """
        extension = str(path).rsplit(".", maxsplit=1)[-1]

        return VideoReader.extension_to_reader.get(extension, OpenCVVideoReader)(path, **kwargs)

    @staticmethod
    def ensure_3d(frame: np.ndarray) -> np.ndarray:
        """Ensure that frame is a 3 dimensional array

        Args:
            frame (np.ndarray): Frame to check

        Returns:
            np.ndarray: Valid frame with 3 dimensions (H, W, C)

        """
        if len(frame.shape) == 2:
            frame = frame[..., None]

        if len(frame.shape) != 3:
            raise ValueError(f"Unhandled dimensions for the given frame: {frame.shape}")

        return frame


class OpenCVVideoReader(VideoReader):
    """Wrapper around opencv VideoCapture

    Default VideoReader when opening a file.

    """

    def __init__(self, path: Union[str, os.PathLike], **kwargs):
        super().__init__(path, **kwargs)
        self.video = cv2.VideoCapture(str(path), **kwargs)
        assert self.video.grab(), "No frames found in the video"
        self.channels = self.retrieve().shape[-1]

        self.fps = int(self.video.get(cv2.CAP_PROP_FPS))
        self.length = int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))
        self.shape = (int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)), int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)))

    def release(self) -> None:
        super().release()
        self.video.release()

    def grab(self) -> bool:
        grabbed = self.video.grab()
        if grabbed:
            self.frame_id += 1
        else:
            self.seek(self.frame_id)
        return grabbed

    def retrieve(self) -> np.ndarray:
        return self.ensure_3d(self.video.retrieve()[1])

    def seek(self, frame_id: int) -> None:
        if not 0 <= frame_id < self.length:
            raise EOFError(f"Seeking outside of video limits: {frame_id} not in [{0}, {self.length}[")
        self.video.set(cv2.CAP_PROP_POS_FRAMES, frame_id + 1)
        self.frame_id = frame_id

    def _check_frame_id(self) -> None:
        assert self.frame_id == self.video.get(cv2.CAP_PROP_POS_FRAMES) - 1


class TiffVideoReader(VideoReader):
    """Special VideoReader for multi images tif files, not supported by opencv

    Uses PIL as backend

    """

    supported_extensions = ["tif", "tiff"]

    def __init__(self, path: Union[str, os.PathLike], **kwargs):
        super().__init__(path, **kwargs)
        self.video = Image.open(path, **kwargs)
        self.channels = self.retrieve().shape[-1]
        self.shape = self.video.size[::-1]
        self.length = self.video.n_frames  # type: ignore

    def release(self) -> None:
        super().release()
        self.video.close()

    def grab(self) -> bool:
        try:
            self.seek(self.frame_id + 1)
            return True
        except EOFError:
            return False

    def retrieve(self) -> np.ndarray:
        frame = self.ensure_3d(np.array(self.video))
        return np.flip(frame, 2)  # In PIL, RGB images are read as RGB. Let's flip it to BGR

    def seek(self, frame_id: int) -> None:
        self.video.seek(frame_id)
        self.frame_id = frame_id

    def _check_backend_frame_id(self) -> None:
        assert self.frame_id == self.video.tell()



from typing import Tuple

import numba  # type: ignore
import numpy as np


@numba.njit()
def _fast_build_links(indices: np.ndarray, shape: Tuple[int, int]):
    """Extract the links from the sorted indices"""
    n = min(shape)
    i_valid = np.full((shape[0],), True)
    j_valid = np.full((shape[1],), True)

    links = []

    for i, j in indices:  # O(nm), but much faster has we stop before having seen most indices
        if i_valid[i] and j_valid[j]:
            i_valid[i] = False
            j_valid[j] = False
            links.append((i, j))

            if len(links) >= n:
                break

    return np.array(links)


def greedy_assignment_solver(dist: np.ndarray, eta: float = np.inf):
    """Solve assignement problem in a greedy way

    Iteratively select the minimum cost, then deleting its row/column.

    Stops when the cost matrix is empty (no more rows or columns) or if the selected
    cost is higher than `eta`

    Args:
        dist (np.ndarray): Distance matrix
                Shape: (N, M), dtype: float
        eta (float): Hard thresholding
            Default: inf (No thresholding)

    Returns:
        np.ndarray: Links (i, j)
            Shape: (L, 2), dtype: uint16
    """
    if min(dist.shape) == 0:
        return np.zeros((0, 2), dtype=np.uint16)

    # Create a sorted list of indices to investigate => O(nm log(nm))
    valid = dist <= eta
    indices = np.indices(dist.shape).transpose((1, 2, 0))[valid]
    indices = indices[np.argsort(dist[valid])]
    if indices.size == 0:  # No links can be done
        return np.zeros((0, 2), dtype=np.uint16)

    links = _fast_build_links(indices, dist.shape[:2])

    return np.array(links, dtype=np.uint16).reshape(-1, 2)  # Ensure uint16 and (L, 2)



import dataclasses
from typing import Optional, Tuple, Union
import warnings

import numpy as np
import torch

import byotrack
from .base import AssociationMethod, FrameByFrameLinker, FrameByFrameLinkerParameters


@dataclasses.dataclass
class NearestNeighborParameters(FrameByFrameLinkerParameters):
    """Parameters of NearestNeighborLinker

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
            Default: 3
        n_gap (int): Number of frames with no association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
        ema (float): Optional exponential moving average to reduce detection noise. Detection positions are smoothed
            using this EMA. Should be smaller than 1.It use: x_{t+1} = ema x_{t} + (1 - ema) det(t)
            As motion is not modeled, EMA may introduce lag that will hinder tracking. It is more effective with
            optical flow to compensate motions, in this case, a typical value is 0.5, to average the previous position
            with the current measured one. For more advanced modelisation, see `KalmanLinker`.
            Default: 0.0 (No EMA)
        fill_gap (bool): Fill the gap of missed detections using a forward optical flow
            propagation (Only when optical flow is provided). We advise to rather use a
            ForwardBackward interpolation using the same optical flow: it will produce
            smoother interpolations.
            Default: False

    """

    def __init__(
        self,
        association_threshold: float,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        ema=0.0,
        fill_gap=False,
    ):
        super().__init__(association_threshold, n_valid, n_gap, association_method)
        self.ema = ema
        self.fill_gap = fill_gap

    ema: float
    fill_gap: bool


class NearestNeighborLinker(FrameByFrameLinker):
    """Frame by frame linker by associating with the closest detections

    Motion is not modeled, but if an optical flow method is provided, it
    will be used to compensate motion online. Matching is done from a simple Euclidean
    distance. This can be easily changed by inheriting this class and overriding the `cost` method.

    See `FrameByFrameLinker` for the other attributes.

    Attributes:
        specs (NearestNeighborParameters): Parameters specifications of the algorithm.
            See `NearestNeighborParameters`.
        active_positions (Optional[torch.Tensor]): The positions of actives tracks,
            if undetected it is estimated by optical flow propagation.
            Shape: (N, D), dtype: float32

    """

    progress_bar_description = "Nearest Neighbor linking"

    def __init__(
        self, specs: NearestNeighborParameters, optflow: Optional[byotrack.OpticalFlow] = None, save_all=False
    ) -> None:
        super().__init__(specs, optflow, save_all)
        self.specs: NearestNeighborParameters
        self.active_positions: Optional[torch.Tensor] = None

        if self.specs.fill_gap and not self.optflow:
            warnings.warn("Optical flow has not been provided. Gap cannot be filled")

    def reset(self) -> None:
        super().reset()
        self.active_positions = None

    def motion_model(self) -> None:
        if self.optflow and self.optflow.flow_map is not None:
            if self.active_positions is not None:
                self.active_positions = torch.tensor(
                    self.optflow.optflow.transform(self.optflow.flow_map, self.active_positions.numpy())
                )

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        if self.active_positions is None:
            self.active_positions = torch.empty((0, detections.position.shape[1]))

        return torch.cdist(self.active_positions, detections.position), self.specs.association_threshold

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, links: torch.Tensor):
        if self.active_positions is None:
            self.active_positions = torch.empty((0, detections.position.shape[1]))

        # Update tracks positions with detections
        # Optionally using an EMA to reduce detections noise
        self.active_positions[links[:, 0]] -= (1.0 - self.specs.ema) * (
            self.active_positions[links[:, 0]] - detections.position[links[:, 1]]
        )

        # Update active track handlers
        active_mask = self.update_active_tracks(links)

        # Create new track handlers for unmatched detections
        unmatched = self.handle_extra_detections(detections, links)

        # Merge still active positions and new ones
        self.active_positions = torch.cat((self.active_positions[active_mask], detections.position[unmatched]))

        self.all_positions.append(self.active_positions.clone())

        if not self.specs.fill_gap:  # Erase optical flow predictions from the stored positions
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] == -1:  # Not truly detected
                    self.all_positions[-1][i, :] = torch.nan



from typing import Any, Callable, Dict, Optional, Union

import numpy as np

from byotrack import OpticalFlow


class SkimageOpticalFlow(OpticalFlow):
    """Wraps Scikit-Image optical flow implementations

    Usage:

    .. code-block:: python

        import skimage.registration  # Requires to install scikit image
        from byotrack.implementation.optical_flow.skimage import SkimageOpticalFlow

        # See the documentation of Scikit-Image for each algorithm to correctly set the parameters
        parameters = {}

        # ILK
        optflow = SkimageOpticalFlow(skimage.registration.optical_flow_ilk, parameters=parameters)

        # TVL1
        optflow = SkimageOpticalFlow(skimage.registration.optical_flow_tvl1, parameters=parameters)


    Attributes:
        method (Callable[Any, np.ndarray]): The optical flow function from skimage. In skimage, only two currently
            exists: skimage.registration.optical_flow_ilk or skimage.registration.optical_flow_tvl1.
        parameters (Dict[str, Any]): Parameters that are given to `method` as known arguments.

    """

    def __init__(
        self,
        method: Callable[..., np.ndarray],
        downscale: Union[float, np.ndarray] = 2,
        blur: Union[None, float, np.ndarray] = None,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(downscale, blur)
        self.method = method
        self.parameters = {} if parameters is None else parameters

    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        # Average channels (multi channel is not supported)
        reference = reference.mean(axis=-1)
        moving = moving.mean(axis=-1)

        # Flow map is already at the right format (2, H', W') with coordinates (i, j)
        return self.method(reference, moving, **self.parameters).astype(np.float32)



from abc import abstractmethod
import dataclasses
from typing import List, Optional, Tuple, Union

import enum

import numpy as np
import pylapy
import torch

import byotrack

from .greedy_lap import greedy_assignment_solver


class AssociationMethod(enum.Enum):
    """Association methods (Greedy or Jonker-Volgenant)

    * GREEDY
        Select the best match between tracks and detections iteratively until
        no match can be selected below the cost limit eta. It is usually not optimal for tracking
        but it is much faster.
    * OPT_HARD
        Solve the linear association problem (see `pylapy`).
        Hard threshold the association matrix with the cost limit eta.
    * OPT_SMOOTH
        Solve a cost_limit extended association problem (see `pylapy`)
        It relaxes the linear association problem, allowing to not link a node
        for the cost limit eta.
    """

    GREEDY = "greedy"
    OPT_HARD = "opt_hard"
    OPT_SMOOTH = "opt_smooth"

    def solve(self, cost: torch.Tensor, eta: float = np.inf) -> torch.Tensor:
        """Solve tracks-to-detections association

        Args:
            cost (torch.Tensor): Cost matrix
                    Shape: (N, M), dtype: float
            eta (float): Cost limit
                Default: inf (No thresholding)

        Returns:
            torch.Tensor: Links (i, j)
                Shape: (L, 2), dtype: int32
        """

        if self == AssociationMethod.OPT_HARD:
            cost[cost > eta] = np.inf
            return torch.tensor(pylapy.LapSolver().solve(cost.numpy()).astype(np.int32))

        if self == AssociationMethod.OPT_SMOOTH:
            return torch.tensor(pylapy.LapSolver().solve(cost.numpy(), eta).astype(np.int32))

        return torch.tensor(greedy_assignment_solver(cost.numpy(), eta).astype(np.int32))


class TrackHandler:
    """Handle a track during the tracking procedure

    It accumulates the track data at each new association and store the optional motion model data.

    A TrackHandler is created for each unlinked detections in the linking process and then updated with
    the following associated detections. At the beginining, the track is considered HYPOTHETICAL.
    For a track to be considered valid, it requires n_valid consecutive associated detections after the track creation
    (state: HYPOTHETICAL => VALID). It a miss detection occurs during this time interval, then the track is deleted
    and considered invalid (state: HYPOTHETICAL => INVALID).

    Once confirmed, a VALID track is resilient to miss detections, waiting n_gap frames before ending the track
    (VALID => FINISHED).

    Attributes:
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
        n_gap (int): Number of frames with no association before the track termination.
        start (int): Starting frame of the track
        identifier (int): Identifier of the track handler (and of the track)
        track_state (TrackState): Current state of the handler
        last_association (int): Number of frames since the last association
        detection_ids (List[int]): Identifiers of the associated detection (-1 if None)
        track_ids (List[int]): Index of the track at each frame in the `linker.active_tracks` list.
            It allows the linker to store data as tensor and be able to rebuild tracks at the end.

    """

    class TrackState(enum.IntEnum):
        """TrackState of a TrackHandler

        * HYPOTHETICAL
            Initial state before validation of the track.
        * VALID
            The track has been validated and is still active
        * FINISHED
            The track is valid and finished
        * INVALID
            The track is not valid and deleted

        """

        HYPOTHETICAL = 0
        VALID = 1
        FINISHED = 2
        INVALID = 3

    def __init__(self, n_valid: int, n_gap: int, start: int, identifier: int) -> None:
        self.n_valid = n_valid
        self.n_gap = n_gap
        self.start = start
        self.identifier = identifier
        self.track_state = TrackHandler.TrackState.HYPOTHETICAL
        self.last_association = 0
        self.detection_ids: List[int] = []
        self.track_ids: List[int] = []

    def __len__(self) -> int:
        return len(self.detection_ids) - self.last_association

    def is_active(self) -> bool:
        return self.track_state < 2

    def update(self, frame_id: int, detection_id: int) -> None:
        """Update track handler. It stores the detection_id and update the track state.

        It should be called for each time frame and each active track.

        Args:
            frame_id (int): The current frame. This is given for safety checks
                to ensure that the Linker and TrackHandler agree.
            detection_id (int): Detection id in the Detections object.
                -1 if not associated to a particular detection.

        """
        assert self.is_active()
        assert len(self.track_ids) == len(
            self.detection_ids
        ), "The linker should call `update` then `register_track_id` at each linking step"
        assert (
            self.start + len(self.detection_ids) == frame_id
        ), "The linker should update each active track on each time frame."

        self.detection_ids.append(detection_id)

        if detection_id == -1:  # Not associated
            self.last_association += 1

            if self.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                self.track_state = TrackHandler.TrackState.INVALID

            if self.last_association > self.n_gap:
                self.track_state = TrackHandler.TrackState.FINISHED
        else:
            self.last_association = 0

            if self.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                if len(self.detection_ids) >= self.n_valid:
                    self.track_state = TrackHandler.TrackState.VALID

    def register_track_id(self, track_id: int) -> None:
        """For still active tracks, it registers the track id after the update step.

        Args:
            track_id (int): The index of the track in `linker.active_tracks` at this time frame.
        """
        self.track_ids.append(track_id)


class OnlineFlowExtractor:
    """Extract optical flow maps online from a video"""

    def __init__(self, optflow: byotrack.OpticalFlow) -> None:
        self.optflow = optflow
        self.flow_map: Optional[np.ndarray] = None
        self.src: Optional[np.ndarray] = None

    def reset(self) -> None:
        """Reset the flow extractor"""
        self.flow_map = None
        self.src = None

    def update(self, frame: np.ndarray) -> None:
        """Extract the flow for a new given frame

        It will compute and register the flow map between the last given frame
        and the current frame.

        Args:
            frame (np.ndarray): Current frame of the video

        """
        dst = self.optflow.preprocess(frame)
        if self.src is None:
            self.src = dst

        self.flow_map = self.optflow.compute(self.src, dst)
        self.src = dst


@dataclasses.dataclass
class FrameByFrameLinkerParameters:  # pylint: disable=too-many-instance-attributes
    """Parameters of the abstract FrameByFrameLinker

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
            Default: 3
        n_gap (int): Number of frames with no association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
    """

    def __init__(
        self,
        association_threshold: float,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
    ):
        self.association_threshold = association_threshold
        self.n_valid = n_valid
        self.n_gap = n_gap
        self.association_method = (
            association_method
            if isinstance(association_method, AssociationMethod)
            else AssociationMethod[association_method.upper()]
        )

    association_threshold: float
    n_valid: int
    n_gap: int
    association_method: AssociationMethod


class FrameByFrameLinker(byotrack.OnlineLinker):
    """Links detections online using frame-by-frame association

    Abstract class for frame-by-frame linker. It decomposes the update step in 5 parts:

    1. Optional optical flow computations (Handled by this class)
    2. Motion modeling to predict track positions (`motion_model`)
    3. Track-to-detection cost computation (`cost`)
    4. Solving the linear association problem (Handled by this class)
    5. Post matching update to handle tracks (`post_association`)

    The association relies on the AssociationMethod enum and tracks handling is done with
    TrackHandler.

    It follows the tracks handling strategy describe in KOFT [9].

    Attributes:
        specs (FrameByFrameLinkerParameters): Parameters specifications of the algorithm.
            See `FrameByFrameLinkerParameters`.
        optflow (Optional[OnlineFlowExtractor]): Optional wrapper around the given optional OpticalFlow that
            will extract flow maps of the video online. (The underlying OpticalFlow object is accessible in
            self.optflow.optflow)
            Default: None
        save_all (bool): Save metadata useless for the final building of tracks
            but that could be usefull for analysis. For instance, it will keep invalid tracks.
        frame_id (int): Current frame id of the linking process
        inactive_tracks (List[TrackHandler]): Terminated tracks
        active_tracks (List[TrackHandler]): Current track handlers
        all_positions (List[torch.Tensor]): Positions of the active tracks at each seen frames.
            Using the valid track handlers `track_ids`, it allows the reconstruction of tracks.

    """

    def __init__(
        self,
        specs: FrameByFrameLinkerParameters,
        optflow: Optional[byotrack.OpticalFlow] = None,
        save_all=False,
    ) -> None:
        super().__init__()
        self.specs = specs
        self.optflow = OnlineFlowExtractor(optflow) if optflow else None
        self.save_all = save_all
        self.frame_id = -1
        self.inactive_tracks: List[TrackHandler] = []
        self.active_tracks: List[TrackHandler] = []
        self.all_positions: List[torch.Tensor] = []

    def reset(self) -> None:
        super().reset()
        if self.optflow:
            self.optflow.reset()
        self.frame_id = -1
        self.inactive_tracks = []
        self.active_tracks = []
        self.all_positions = []

    def collect(self) -> List[byotrack.Track]:
        tracks = []
        for handler in self.inactive_tracks + self.active_tracks:
            if handler.track_state in (handler.TrackState.INVALID, handler.TrackState.HYPOTHETICAL):
                continue  # Ignore non-valid tracks

            points = torch.cat(
                [
                    positions[track_id : track_id + 1]
                    for track_id, positions in zip(
                        handler.track_ids[: len(handler)], self.all_positions[handler.start :]
                    )
                ]
            )

            tracks.append(
                byotrack.Track(
                    handler.start,
                    points,
                    handler.identifier,
                    torch.tensor(handler.detection_ids[: len(handler)], dtype=torch.int32),
                )
            )
        return tracks

    @abstractmethod
    def motion_model(self) -> None:
        """Optional modelisation of motion for tracks

        It can be used to update some internal state of the tracker after the optical flow computation
        and before the distance computation.
        """

    @abstractmethod
    def cost(self, frame: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        """Compute the association cost between active tracks and detections

        It also returns the threshold to use (Depending on the dist you use, association_threshold
        could be related to a more meaning full quantity than the cost itself).
        For instance, when using a squared Euclidean distance, the association threshold could be
        express as the distance in pixel, and this function could square it.
        For likelihood association, you could provide the association threshold as a probability
        and use -log(threshold) as the true threshold. (See `KalmanLinker` and `NearestNeighborLinker`)

        Args:
            frame (np.ndarray): The current frame of the video
                Shape: (H, W, C), dtype: float
            detections (byotrack.Detections): Detections for the given frame

        Returns:
            torch.Tensor: The cost matrix between active tracks and detections
                Shape: (n_tracks, n_dets), dtype: float32
            float: The association threshold to use.
                It can be different than `self.association_threshold` depeding on the dist build here

        """

    @abstractmethod
    def post_association(self, frame: np.ndarray, detections: byotrack.Detections, links: torch.Tensor):
        """Update the tracks and the internal variables of the tracker

        It should call the `update` method of each active tracks and update any internal model/data.
        It should also create new track handlers for each extra detection.
        Finally, it is also responsible to register the position of each active track in `all_positions`
        for the current time frame.

        Args:
            frame (np.ndarray): The current frame of the video
                Shape: (H, W, C), dtype: float
            detections (byotrack.Detections): Detections for the given frame
            links (torch.Tensor): The links made between active tracks and the detections
                Shape: (L, 2), dtype: int32

        """

    def update_active_tracks(self, links: torch.Tensor) -> torch.Tensor:
        """Calls `update` for active tracks and return a boolean mask that indicates which track is still active

        Tracks that are terminated are stored inside `inactive_tracks` and dropped from `active_tracks`.

        It can be called inside `post_association` to simplify the code.

        Args:
            links (torch.Tensor): The links made between active tracks and the detections
                Shape: (L, 2), dtype: int32

        Returns:
            torch.Tensor: Boolean tensor indicating True for still active tracks

        """
        i_to_j = torch.full((len(self.active_tracks),), -1, dtype=torch.int32)
        i_to_j[links[:, 0]] = links[:, 1]
        active_mask = torch.full((len(self.active_tracks),), False)
        still_active = []
        for i, track in enumerate(self.active_tracks):
            track.update(self.frame_id, int(i_to_j[i].item()))

            # Check if track is still active
            if track.is_active():
                still_active.append(track)
                active_mask[i] = True
            elif track.track_state == TrackHandler.TrackState.FINISHED or self.save_all:
                self.inactive_tracks.append(track)

        self.active_tracks = still_active

        return active_mask

    def handle_extra_detections(self, detections: byotrack.Detections, links: torch.Tensor) -> torch.Tensor:
        """Handle extra detections by creating new track handlers

        It can be called inside `post_association` to create track handlers from extra detections. It will
        return the extra detections positions and ids to be further used by `post_association`.

        Args:
            detections (byotrack.Detections): Detections for the given frame
            links (torch.Tensor): The links made between active tracks and the detections
                Shape: (L, 2), dtype: int32

        """
        # Find unmatched measures
        unmatched = torch.full((len(detections),), True)
        unmatched[links[:, 1]] = False
        unmatched_ids = torch.arange(len(detections))[unmatched]
        # unmatched_measures = detections.position[unmatched]

        # Create a new active track for each unmatched measure
        for i in range(unmatched_ids.shape[0]):
            handler = TrackHandler(
                self.specs.n_valid, self.specs.n_gap, self.frame_id, len(self.inactive_tracks) + len(self.active_tracks)
            )
            handler.update(self.frame_id, int(unmatched_ids[i]))
            self.active_tracks.append(handler)

        return unmatched

    def update(self, frame: np.ndarray, detections: byotrack.Detections) -> None:
        self.frame_id += 1

        # Compute the flow map if optflow given
        if self.optflow:
            self.optflow.update(frame)

        self.motion_model()
        cost, threshold = self.cost(frame, detections)
        links = self.specs.association_method.solve(cost, threshold)

        self.post_association(frame, detections, links)

        assert len(self.all_positions[-1]) == len(
            self.active_tracks
        ), "Wrong implementation of post_association. It should register the positions of each active tracks"
        for i, track in enumerate(self.active_tracks):
            track.register_track_id(i)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Sequence, Union
import warnings

import numpy as np
import tqdm.auto as tqdm

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class Linker(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base class for linkers in videos

    Link detections through time to build tracks.

    Each linker can define a set of parameters (See `ParametrizedObjectMixin`)

    """

    @abstractmethod
    def run(
        self, video: Union[Sequence[np.ndarray], np.ndarray], detections_sequence: Sequence[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        """Run the linker on a whole video

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Sequence of T frames (array).
                Each array is expected to have a shape (H, W, C)
            detections_sequence (Sequence[byotrack.Detections]): Detections for each frame
                Detections is expected for each frame of the video, in the same order.
                (Note that for a given frame, the Detections can be empty)

        Returns:
            Collection[byotrack.Track]: Tracks of particles

        """


class OnlineLinker(Linker):
    """Online linker, it tracks particles one frame at a time

    The linking algorithm is dividing into three steps:

    * reset: Reset the algorithm
    * update: Update the algorithm with a new frame and detections
    * collect: Collect the constructed tracks up to the last frame

    """

    progress_bar_description = "Linking"

    def reset(self) -> None:
        """Reset the linking algorithm

        Flush all data stored from a previous linking and prepare a new linking
        """

    @abstractmethod
    def update(self, frame: np.ndarray, detections: byotrack.Detections) -> None:
        """Progress in the linking step by one frame

        Will update the internal algorithm by a single frame and its detections.

        Args:
            frame (np.ndarray): Frame of the video
                Shape: (H, W, C), dtype: float
            detections (byotrack.Detections): Detections for the given frame

        """

    @abstractmethod
    def collect(self) -> Collection[byotrack.Track]:
        """Collect and build all the tracks up to the last given frame

        Returns:
            Collection[byotrack.Track]: Tracks from the last reset to the last given frame.

        """

    def run(
        self, video: Union[Sequence[np.ndarray], np.ndarray], detections_sequence: Sequence[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        if len(video) != len(detections_sequence):
            warnings.warn(
                f"""Expected to have one Detections for each frame of the video.

            There are {len(detections_sequence)} Detections for {len(video)} frames.
            This can lead to unexpected behavior. By default we assume that the first Detections
            is aligned with the first frame and stop when the end of shortest sequence is reached.
            """
            )
        self.reset()

        progress_bar = tqdm.tqdm(desc=self.progress_bar_description, total=min(len(video), len(detections_sequence)))

        for frame, detections in zip(video, detections_sequence):
            self.update(frame, detections)
            progress_bar.update()

        progress_bar.close()

        return self.collect()



"""ByoTrack

Unified python API for biological particle tracking.

We provide a unified python API for tracking that can be easily
extended with new (and old) algorithms. We also provide
implementations of well-known algorithms following our API.

Overview:
* Video
    * Able to read classical format (supported by opencv) + tiff
* Particle Tracking
    * MultiStepTracker (Detect / Link / Refine)
* Particle Detections
    * Wavelet Detector [2] (Similar as the one in Icy [1] but coded in pytorch)
    * Stardist [3] (Inference only. Training should be done with the official implementation)
* Particle Linking
    * EMHT [4] (Wraps the implementation in Icy [1], requires Icy to be installed)
    * u-track / TrackMate [7] (Wraps TrackMate [6, 8] implementation in ImageJ/Fiji, requires Fiji to be installed)
* Tracks Refining
    * Cleaning
    * EMC2 [5]: Track stitching (gap closing)
    * Interpolate missing positions



Getting started:
----------------

```python
import byotrack

# Load some specific implementations
from byotrack.implementation.detector.wavelet import WaveletDetector
from byotrack.implementation.linker.icy_emht import IcyEMHTLinker
from byotrack.implementation.refiner.cleaner import Cleaner
from byotrack.implementation.refiner.stitching import EMC2Stitcher

# Read a video from a path, normalize and aggregate channels
video = byotrack.Video(video_path)
transform_config = VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
video.set_transform(transform_config)

# Create a multi step tracker
## First the detector
## Smaller scale <=> search for smaller spots
## The noise threshold is linear with k. If you increase it, you will retrieve less spots.
detector = WaveletDetector(scale=1, k=3.0, min_area=5)

## Second the linker
## Hyperparameters are automatically chosen by Icy
linker = IcyEMHTLinker(icy_path)

## Finally refiners
## If needed you can add Cleaning and Stitching operations
refiners = []
if True:
    refiners.append(Cleaner(5, 3.5))  # Split tracks on position jumps and drop small ones
    refiners.append(EMC2Stitcher())  # Merge tracks if they track the same particle

tracker = byotrack.MultiStepTracker(detector, linker, refiners)

# Run the tracker
tracks = tracker.run(video)

# Save tracks
byotrack.Track.save(tracks, output_path)
```

Please refer to the ![official documentation](https://byotrack.readthedocs.io/en/latest/?badge=latest)

"""

# Import API and main features but no implementation
from byotrack.api.detector.detections import Detections
from byotrack.api.detector.detector import BatchDetector, Detector
from byotrack.api.linker import Linker, OnlineLinker
from byotrack.api.optical_flow.optical_flow import OpticalFlow
from byotrack.api.refiner import Refiner
from byotrack.api.tracker import BatchMultiStepTracker, MultiStepTracker, Tracker
from byotrack.api.tracks import Track
from byotrack.video import Video, VideoTransformConfig


__version__ = "0.7.1"



from __future__ import annotations

import os
from typing import cast, Collection, Optional, Sequence, Tuple, Union

import numpy as np
import torch

import byotrack  # pylint: disable=cyclic-import


def _check_points(points: torch.Tensor) -> None:
    """Check points validity (type, shape)"""
    assert len(points.shape) == 2
    assert points.shape[0] > 0
    assert points.shape[1] in (2, 3)  # Only 2d and 3d supported
    assert points.dtype is torch.float32


def _check_detection_ids(detection_ids: torch.Tensor, size: int) -> None:
    assert detection_ids.shape == (size,)
    assert detection_ids.dtype is torch.int32


class Track:
    """Track for a given particle

    A track is defined by an (non-unique) identifier, a starting frame and a succession of positions.
    In a detect-then-track context, a track can optionally contains the detection identifiers
    for each time frame (-1 if non-linked to any particular detection at this time frame)

    Attributes:
        identifier (int): Identifier of the track (non-unique)
        start (int): Starting frame of the track
        points (torch.Tensor): Positions (i, j) of the particle (from starting frame to ending frame)
            Shape: (T, D), dtype: float32
        detection_ids (torch.Tensor): Detection id for each time frame (-1 if unknown or non-linked
            to a particular detection at this time frame)
            Shape: (T,), dtype: int32

    """

    _next_identifier = 0

    def __init__(
        self,
        start: int,
        points: torch.Tensor,
        identifier: Optional[int] = None,
        detection_ids: Optional[torch.Tensor] = None,
    ) -> None:
        if identifier is None:
            self.identifier = Track._next_identifier
            Track._next_identifier += 1
        else:
            self.identifier = identifier

        if detection_ids is None:  # All are unknown (-1)
            detection_ids = torch.full((len(points),), -1, dtype=torch.int32)

        _check_points(points)
        _check_detection_ids(detection_ids, len(points))

        self.start = start
        self.points = points
        self.detection_ids = detection_ids

    def __len__(self) -> int:
        return self.points.shape[0]

    def __getitem__(self, frame_id: int) -> torch.Tensor:
        """Return the position of tracked particle for the given frame_id

        Args:
            frame_id (int): id of the frame in the video

        Returns:
            torch.Tensor: Position (i, j) at the given frame (NaN if unknown)
                Shape: (D, ), dtype: float32

        """
        if self.start <= frame_id < self.start + len(self.points):
            return self.points[frame_id - self.start]

        return torch.full((self.points.shape[1],), torch.nan)

    def overlaps_with(self, other: Track, tolerance=0) -> bool:
        """Test if this track overlaps with another one in time.

        Args:
            other (Track): The other track
            tolerance (int): Time tolerance before detecting an overlap.
                Default: 0 (no tolerance)

        """
        return self.start < other.start + len(other) - tolerance and self.start + len(self) - tolerance > other.start

    @staticmethod
    def tensorize(tracks: Collection[Track], frame_range: Optional[Tuple[int, int]] = None) -> torch.Tensor:
        """Convert a collection of tracks into a tensor on a given frame_range

        Useful view of the data to speedup some mathematical operations

        Args:
            tracks (Collection[Track]): A collection of tracks (usually from the same video)
            frame_range (Optional[Tuple[int, int]]): Frame range (start included, end excluded)
                If None, the minimal range to hold all points is computed

        Returns:
            torch.Tensor: Tracks data in a single tensor
                Shape: (T, N, D), dtype: float32

        """
        if not tracks:
            raise ValueError("Cannot tensorize an empty collection of Tracks")

        # Find the spatial dimension (all tracks should share the same one)
        dim = next(iter(tracks)).points.shape[1]

        # Compute once start and end of each track
        starts = [track.start for track in tracks]
        ends = [track.start + len(track) for track in tracks]

        if frame_range is None:
            start = min(starts)
            end = max(ends)
        else:
            start, end = frame_range

        points = torch.full((end - start, len(tracks), dim), torch.nan)

        for i, (track_start, track_end, track) in enumerate(zip(starts, ends, tracks)):
            if track_end <= start or track_start >= end:
                continue

            track_end = min(end, track_end)

            points[max(0, track_start - start) : track_end - start, i, :] = track.points[
                max(0, start - track_start) : track_end - track_start
            ]

        return points

    @staticmethod
    def _tensorize_det_ids(tracks: Collection[Track]) -> torch.Tensor:
        """Build a detection identifiers tensor for multiple tracks

        Args:
            tracks (Collection[Track]): A collection of tracks (usually from the same video)

        Returns:
            torch.Tensor: Detection identifiers in a single tensor
                Shape: (T, N), dtype: int32
        """
        if not tracks:
            raise ValueError("Cannot tensorize an empty collection of Tracks")

        # Compute once start and end of each track
        starts = [track.start for track in tracks]
        ends = [track.start + len(track) for track in tracks]

        start = min(starts)
        end = max(ends)
        ids = torch.full((end - start, len(tracks)), -1, dtype=torch.int32)

        for i, (track_start, track_end, track) in enumerate(zip(starts, ends, tracks)):
            ids[track_start - start : track_end - start, i] = track.detection_ids

        return ids

    @staticmethod
    def save(tracks: Collection[Track], path: Union[str, os.PathLike]) -> None:
        """Save a collection of tracks to path

        Format: pt (pytorch)

        .. code-block::

            {
                "offset": int
                "ids": Tensor (N, ), int64
                "points": Tensor (T, N, D), float32
                "det_ids": Tensor (T, N), int32
            }

        Args:
            tracks (Collection[Track]): Tracks to save
            path (str | os.PathLike): Output path

        """
        ids = torch.tensor([track.identifier for track in tracks])
        offset = min(track.start for track in tracks)
        points = Track.tensorize(tracks)
        det_ids = Track._tensorize_det_ids(tracks)
        torch.save({"offset": offset, "ids": ids, "points": points, "det_ids": det_ids}, path)

    @staticmethod
    def load(path: Union[str, os.PathLike]) -> Collection[Track]:
        """Load a collection of tracks from path

        Args:
            path (str | os.PathLike): Input path

        """
        data: dict = torch.load(path, map_location="cpu")
        offset: int = data.get("offset", 0)
        points: torch.Tensor = data["points"]
        ids: torch.Tensor = data["ids"]
        det_ids: torch.Tensor = data.get("det_ids", torch.full(points.shape[:2], -1, dtype=torch.int32))

        frames = torch.arange(points.shape[0])

        tracks = []

        for i, identifier in enumerate(ids.tolist()):
            track_points = points[:, i]
            defined = ~torch.isnan(track_points).all(dim=-1)
            track_points = track_points[defined]
            track_det_ids = det_ids[:, i][defined]
            start = cast(int, frames[defined].min().item())
            tracks.append(Track(start + offset, track_points, identifier, track_det_ids))

        return tracks


def update_detection_ids(  # pylint: disable=too-many-locals
    tracks: Collection[Track], detections_sequence: Sequence[byotrack.Detections], using_segmentation=True
) -> None:
    """Update the `detections_ids` attribute of each track inplace

    For each frame and each track, a perfectly matching detection is searched (the track position should be equal
    to the detection position). If a match is found, it is registered in the `detections_ids` attribute.

    This is useful to fill the `detection_ids` attributes after a wrapping linking code (See EMHT or TrackMate).
    For this code to work, the linking algorithm that produces tracks should use the detection position
    as the track position without using any temporal/spatial smoothing.

    Args:
        tracks (Collection[Track]): The tracks to update inplace
        detections_sequence (Sequence[byotrack.Detections]): Detections for the different frames
            It should directly be the detections used in the linking algorithm
        using_segmentation (bool): Whether to use the segmentation to compute position of detections
            or use position if available. (Icy and Fiji are only given the segmentation)

    """
    frame_range = (min(track.start for track in tracks), max(track.start + len(track) for track in tracks))
    points = Track.tensorize(tracks, frame_range=frame_range)

    # Store tracks as an array of objects to access advance slicing
    # And reset detection_ids
    tracks_array = np.empty((len(tracks),), dtype=np.object_)
    for i, track in enumerate(tracks):
        tracks_array[i] = track
        track.detection_ids[:] = -1

    for frame_id, detections in enumerate(detections_sequence):
        if detections.length == 0:
            continue
        if frame_id < frame_range[0] or frame_id >= frame_range[1]:
            continue

        valid_tracks = ~torch.isnan(points[frame_id - frame_range[0]]).any(dim=1)
        if valid_tracks.sum() == 0:
            continue

        valid_points = points[frame_id - frame_range[0]][valid_tracks]

        # If using seg, we rely solely on the segmentation
        if using_segmentation and "position" in detections.data:
            detections = byotrack.Detections({"segmentation": detections.segmentation})

        # Compute dist between tracks and detections at time t (Shape: (N_det, N_track))
        dist = (valid_points[None] - detections.position[:, None]).abs().sum(dim=-1)
        mini, argmin = torch.min(dist, dim=0)
        match = mini < 1e-5  # Keep only perfect matches
        argmin = argmin[match]

        for i, track in enumerate(tracks_array[valid_tracks.numpy()][match.numpy()]):
            track.detection_ids[frame_id - track.start] = argmin[i]



from __future__ import annotations

import numpy as np


# pylint: disable=too-few-public-methods


class ChannelSelect:
    """Select a given channel

    Attributes:
        channel (int): Channel to keep (0, 1 or 2)

    Args:
        frame (np.ndarray): Frame of the video
            Shape: (..., H, W, C)

    Returns:
        np.ndarray: Filtered frame with a single channel
            Shape: (..., H, W, 1)

    """

    def __init__(self, channel: int) -> None:
        """Constructor

        Args:
            channel (int): Selected channel

        """
        self.channel = channel

    def __call__(self, frame: np.ndarray) -> np.ndarray:
        return frame[..., self.channel : self.channel + 1]


class ChannelAvg:
    """Average channels into a single one

    Args:
        frame (np.ndarray): Frame of the video
            Shape: (..., H, W, C)

    Returns:
        np.ndarray: Average of channels
            Shape: (..., H, W, 1)

    """

    def __call__(self, frame: np.ndarray) -> np.ndarray:
        return np.mean(frame, axis=-1, keepdims=True)


class ScaleAndNormalize:
    """Scale and Normalize each channel into [0, 1]

    min and max values are computed using quantile of the video to improve stability

    Attributes:
        q_min (float): Quantile of the minimum value to consider
        q_max (float): Quantile of the maximum value to consider
        mini (np.ndarray): Minimum value kept (one for each channel)
            Shape: (C, )
        maxi (np.ndarray): Maximum value kept (one for each channel)
            Shape: (C, )
        smooth_clip (float): Smoothness of the clipping process
            If 0, values are clipped on mini/maxi
            Else, values above maxi are log clipped:
            v = 1 + a log((v - 1)/a + 1) for v > 1, with a the smooth_clip factor
            Typical values are between 0 and 1.
            Default: 0 (hard clipping)
        max (np.ndarray): True maximum values (one for each channel) when using smooth clipping
            Shape: (C, )

    Args:
        frame (np.ndarray): Frame of the video
            Shape: (..., H, W, C)

    Returns:
        np.ndarray: Normalized version of the frame in [0, 1]
            Shape: (..., H, W, C)

    """

    # Do not use all the frames of a video because it is both time and memory expensive
    max_frames_for_stats = 100

    def __init__(self, q_min: float, q_max: float, smooth_clip: float = 0) -> None:
        self.q_min = q_min
        self.q_max = q_max
        self.mini = np.array([0.0])
        self.maxi = np.array([1.0])
        self.smooth_clip = smooth_clip
        self.max = np.array([1.0])

    def update_stats(self, frames: np.ndarray) -> None:
        """Update mini and maxi values based on the given frames

        Args:
            frames (np.ndarray): Several frames of the same video to compute the stats
                Shape: (N, H, W, C)

        """
        frames = frames[: self.max_frames_for_stats]
        self.mini = np.quantile(frames, self.q_min, axis=(0, 1, 2))
        self.maxi = np.quantile(frames, self.q_max, axis=(0, 1, 2))

        if self.smooth_clip > 0:
            ratio = frames.max(axis=(0, 1, 2)) / (self.maxi + (self.maxi == 0))
            self.max = 1 + 0.5 * np.log(np.maximum(1, 1 + (ratio - 1) / 0.5))

    def __call__(self, frame: np.ndarray) -> np.ndarray:
        if self.smooth_clip <= 0:  # No smooth clip
            frame = np.clip(frame, self.mini, self.maxi)
            frame -= self.mini
            frame /= self.maxi - self.mini + (self.maxi == self.mini)  # Divide by one if mini == maxi
            return frame

        frame = (frame - self.mini) / (self.maxi - self.mini + (self.maxi == self.mini))
        # Log cliping high values
        frame[frame > 1] = 1 + self.smooth_clip * np.log((frame[frame > 1] - 1) / self.smooth_clip + 1)
        np.clip(frame, 0, self.max, frame)
        return frame / self.max



from typing import Collection, Sequence, Union

import numba  # type: ignore
import numpy as np

import byotrack

from . import dist_stitcher
from .. import propagation


@numba.njit(parallel=True)
def _fast_emc2_dist(propagation_matrix: np.ndarray, skip_mask: np.ndarray, ranges: np.ndarray) -> np.ndarray:
    """Fast implementation of EMC2 distance

    Compute the mininum distance between each track propagation (only in the temporal gap between them)

    Args:
        propagation_matrix (np.ndarray): Tracks data with propagation
            Shape: (T, N, D), dtype: np.float32
        skip_mask (np.ndarray): Skip computation if True
            Shape: (N, N), dtype: bool
        ranges (np.ndarray): Range for each track (start, end)
            Shape: (N, 2), dtype: int
    """
    n_tracks = propagation_matrix.shape[1]
    emc_dist = np.full((n_tracks, n_tracks), np.inf, dtype=np.float32)

    for i in numba.prange(n_tracks):  # pylint: disable=not-an-iterable
        # Could set some verbosity (without parallel)
        # with numba.objmode():
        #     print(f"Computing emc2 dist {i+1} / {n_tracks}", flush=True, end="\r")
        for j in numba.prange(n_tracks):  # pylint: disable=not-an-iterable
            if skip_mask[i, j]:
                continue

            start = ranges[i, 1] - 1  # End of i (taken)
            end = ranges[j, 0]  # Start of j (taken)
            if end < start:
                start, end = end, start

            diff = propagation_matrix[start : end + 1, i] - propagation_matrix[start : end + 1, j]
            emc_dist[i, j] = (diff**2).sum(axis=-1).min()

    return np.sqrt(emc_dist)


class EMC2Stitcher(dist_stitcher.DistStitcher):
    """Implements Elastic Motion Correction and Concatenation (EMC2)

    Stitch tracks using motion corrected positions following the paper:
    T. Lagache, A. Hanson, J. Perez-Ortega, et al., “Tracking calcium dynamics from
    individual neurons in behaving animals”, PLoS computational biology, vol. 17, pp. e1009432, 10 2021.

    Attributes:
        method (str | DirectedPropagate): Method to use for position propagation. Either "constant", "tps"
            or "flow" (see `constant_directed_propagate`, `tps_directed_propagate` or `optical_flow_directed_propagate`)
            or a user defined Callable following the `DirectedPropagate` protocol.
            Default: "tps"
        eta (float): Soft threshold in LAP solving (See `DistStitcher`)
            Default: 5.0 (Pixels)
        kwargs (Dict[str, Any]): Additional parameters given to `method`.
            For instance for "tps", you can set `alpha` (float) which is the regularization parameter of the
            interpolation (see `tps_propation` module and `torch_tps` librarie). We advise to use alpha > 5.0
            (alpha = 10 by default). For "flow" propagation, you should provide `optflow` (byotrack.OpticalFlow)
            to correctly estimate optical flows.
        max_overlap (int): Cannot stitch tracks that overlap more than `max_overlap`
            Default: 5
        max_dist (float): Cannot stich track i and track j if the last position of i and
            first position of j are farther than `max_dist` (ignored if max_dist <= 0)
            Default: 100.0
        max_gap (int): Cannot stich track i and track j if i ended more
            than `max_gap` frame before j started (ignored if max_gap <= 0)
            Default: 250

    """

    def __init__(self, method: Union[str, propagation.DirectedPropagate] = "tps", eta: float = 5.0, **kwargs) -> None:
        super().__init__(self.compute_dist, eta)
        self.max_overlap = 5
        self.max_dist = 100.0
        self.max_gap = 250
        self.method = method
        self.kwargs = kwargs

    def compute_dist(
        self, video: Union[Sequence[np.ndarray], np.ndarray], tracks: Collection[byotrack.Track]
    ) -> np.ndarray:
        """Compute EMC2 distance between tracks

        Compute the mininum distance between each track propagation (only in the temporal gap between them).

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Video of the tracked particles
            tracks (Collection[Track]): Input tracks

        Returns:
            np.ndarray: Distance between each track. (Contains np.inf)
                Shape: (N, N), dtype: float32
        """
        start = min(track.start for track in tracks)
        end = max(track.start + len(track) for track in tracks)

        skip_mask = self.skip_computation(tracks, self.max_overlap, self.max_dist, self.max_gap)
        propagation_matrix = propagation.forward_backward_propagation(
            byotrack.Track.tensorize(tracks), video[start:end], self.method, **self.kwargs
        )
        ranges = np.array([(track.start, track.start + len(track)) for track in tracks])
        ranges -= ranges.min()  # The track tensor has an offset of ranges.min() (Usually 0)

        return _fast_emc2_dist(propagation_matrix.numpy(), skip_mask.numpy(), ranges)



from __future__ import annotations

import dataclasses
import os
from typing import overload, Optional, Sequence, Tuple, Union

import numpy as np

from .reader import VideoReader
from .transforms import ChannelSelect, ChannelAvg, ScaleAndNormalize


@dataclasses.dataclass
class VideoTransformConfig:
    """Configuration for video transformations

    Attributes:
        aggregate (bool): Aggregate channels
        normalize (bool): Scale and Normalize the video in [0, 1]
        selected_channel (Optional[int]): Channel to use for aggregation
            If None, channel average is done. If any, it performs channel selection
        q_min (float): Minimum quantile to use when scaling the video
        q_max (float): Maximum quantile to use when scaling the video
        smooth_clip (float): Smoothness of the clipping process (log clipping)
            See `ScaleAndNormalize`: it logs clip the highest values on q_max.
            If 0.0, hard clipping is done.
        compute_stats_on (int): Number of frames to use to compute the quantiles.

    """

    aggregate: bool = False
    normalize: bool = False
    selected_channel: Optional[int] = None
    q_min: float = 0.0
    q_max: float = 1.0
    smooth_clip: float = 0.0
    compute_stats_on: int = 50


class Video(Sequence[np.ndarray]):
    """Video: Iterable, indexable and sliceable sequence of frames wrapping a VideoReader

    It wraps VideoReader in order to add video transformation (Channel Aggregation, Scaling, Normalization)
    and to add useful pythonic protocols (Sliceable, Indexable, Iterable).

    Return images in BGR by default like opencv as frames are mostly used with opencv afterwards for display.
    Can also return grayscale image (H, W, 1)

    Example:
        .. code-block:: python

            import byotrack

            # Read a video (Usually BGR)
            video = byotrack.Video(video_path)

            # Add a transform that will aggregate channel and normalize in [0, 1] the intensities
            transform_config = byotrack.VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
            video.set_transform(transform_config)

            # Iterate through the video
            for frame in video:
                pass

            # Temporal slicing
            sliced = video[10:50:3]  # Take one frame every three from frame 10 to frame 50.

            # Spatial slicing
            sliced = video[:, 100:200, 150:250]  # All frames on the roi (100:200 x 150:250)


    Attributes:
        shape (Tuple[int, ...]): Shape of the video (Time, Height, Width[, Depth])
        channels (int): Number of channels
        reader (byotrack.VideoReader): Underlying video reader

    """

    def __init__(self, data_source: Union[str, os.PathLike, VideoReader]) -> None:
        super().__init__()

        if isinstance(data_source, (str, os.PathLike)):
            self.reader = VideoReader.open(data_source)
        else:
            self.reader = data_source

        self._slices: Tuple[slice, ...] = tuple(slice(None) for _ in (self.reader.length, *self.reader.shape))
        self._channel_aggregator: Optional[Union[ChannelAvg, ChannelSelect]] = None
        self._normalizer: Optional[ScaleAndNormalize] = None

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(
            slice_length(slice_, shape) for slice_, shape in zip(self._slices, (self.reader.length, *self.reader.shape))
        )

    @property
    def channels(self) -> int:
        return self.reader.channels if self._channel_aggregator is None else 1

    def set_transform(self, transform_config: VideoTransformConfig) -> None:
        """Set the transform (channel_selector and normalizer)

        Args:
            transform_config (byotrack.VideoTransformConfig): Configuration of the transformations

        """
        self._channel_aggregator = None
        if transform_config.aggregate:
            if transform_config.selected_channel is not None:
                self._channel_aggregator = ChannelSelect(transform_config.selected_channel)
            else:
                self._channel_aggregator = ChannelAvg()

        self._normalizer = None
        if transform_config.normalize:
            frames = np.asarray(self[: min(transform_config.compute_stats_on, ScaleAndNormalize.max_frames_for_stats)])
            self._normalizer = ScaleAndNormalize(
                transform_config.q_min, transform_config.q_max, transform_config.smooth_clip
            )
            self._normalizer.update_stats(frames)

    def transform(self, frame: np.ndarray) -> np.ndarray:
        """Transform a frame using channel aggregation and normalization"""
        if self._channel_aggregator:
            frame = self._channel_aggregator(frame)
        if self._normalizer:
            frame = self._normalizer(frame)

        return frame

    def __len__(self) -> int:
        return slice_length(self._slices[0], self.reader.length)  # length of temporal slice

    @overload
    def __getitem__(self, index: int) -> np.ndarray: ...

    @overload
    def __getitem__(self, slice_: slice) -> Video: ...

    @overload
    def __getitem__(self, slices: Tuple[slice, ...]) -> Video: ...

    def __getitem__(self, key):  # pylint: disable=too-many-branches
        """Indexing and slicing operations

        When indexed, it returns the ith frame in the slice
        When sliced, it duplicates the video (wrapper) with the right slice

        Args:
            key (int | slice | Tuple[slice, ...]): index or slice of the video

        Returns:
            np.ndarray | Video: Frame at index or a shallow copy of the video with the right slice

        """
        if isinstance(key, int):
            start, _, step = self._slices[0].indices(self.reader.length)

            if key < 0:
                key += len(self)

            if key < 0 or key >= len(self):
                raise IndexError(f"Index {key} out of range")

            frame_id = start + key * step

            if frame_id == self.reader.frame_id:
                pass  # Skip expensive seek
            if frame_id == self.reader.frame_id + 1:
                if not self.reader.grab():  # Much faster for cv2: Allows a fast frame by frame reading
                    raise RuntimeError(f"Unable to grab frame {frame_id}")
            else:
                try:
                    self.reader.seek(frame_id)
                except EOFError:
                    raise RuntimeError(f"Unable to seek frame {frame_id}") from None

            return self.transform(self.reader.retrieve()[self._slices[1:]])

        if isinstance(key, slice):
            key = (key,)

        if isinstance(key, tuple):
            if len(key) > len(self._slices):
                raise IndexError("Too many indices for video. Only support 3 dimensions slicing (time, height, width).")

            slices = list(self._slices)
            shapes = (self.reader.length, *self.reader.shape)
            for i, slice_ in enumerate(key):
                if not isinstance(slice_, slice):
                    raise TypeError("Unsupported index for Video. Supports only int, slice and Tuple[slice, ...]")

                slices[i] = compose_slice(self._slices[i], slice_, shapes[i])

            # Duplicate
            other = Video(self.reader)
            other._channel_aggregator = self._channel_aggregator
            other._normalizer = self._normalizer
            other._slices = tuple(slices)

            return other

        raise TypeError("Unsupported index for Video. Supports only int, slice and Tuple[slice, ...]")


def compose_slice(slice_1: slice, slice_2: slice, length: int) -> slice:
    """Compose two slices in the given order"""
    # Compute start, stop, step and length of the first slice
    start_1, _, step_1 = slice_1.indices(length)
    length_1 = slice_length(slice_1, length)

    # Same for the second slice
    start_2, stop_2, step_2 = slice_2.indices(length_1)
    length = slice_length(slice_2, length_1)

    # Compute the final start, stop and step
    start = start_1 + start_2 * step_1
    stop = start_1 + stop_2 * step_1
    step = step_1 * step_2

    if start > stop:
        if stop < 0:  # (5, -1, -1) is empty, instead use (5, None, -1)
            return slice(start, None, step)

    return slice(start, stop, step)


def slice_length(slice_: slice, shape: int) -> int:
    """Compute the number of element in a slice"""
    start, stop, step = slice_.indices(shape)
    return max((stop - start - (step > 0) + (step < 0)) // step + 1, 0)



# pylint: disable=duplicate-code

import dataclasses
from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch_kf
import torch_kf.ckf

import byotrack

from .base import AssociationMethod, OnlineFlowExtractor
from .kalman_linker import Cost, KalmanLinker, KalmanLinkerParameters, TrackBuilding


@dataclasses.dataclass
class KOFTLinkerParameters(KalmanLinkerParameters):
    """Parameters of KOFTLinker

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
        detection_std (Union[float, torch.Tensor]): Expected measurement noise on the detection process.
            The detection process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension). See `torch_kf.ckf.constant_kalman_filter`.
            Default: 3.0 pixels
        flow_std (Union[float, torch.Tensor]): Expected measurement noise on the optical flow process.
            The optical flow process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension).
            Default: 1.0 pixels
        process_std (Union[float, torch.Tensor]): Expect process noise. See `torch_kf.ckf.constant_kalman_filter`, the
            process is modeled as constant order-th derivative motion. This quantify how much the supposely "constant"
            order-th derivative can change between two consecutive frames. A common rule of thumb is to use
            3 * process_std ~= max_t(| x^(order)(t) - x^(order)(t+1)|). It can be provided for each dimension).
            Default: 1.5 pixels / frame^order
        kalman_order (int): Order of the Kalman filter to use. 0 is not supported.
            1 for directed brownian motion, 2 for accelerated brownian motions, etc...
            Default: 1
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
            Default: 3
        n_gap (int): Number of frames with no association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
        cost_method (CostMethod): The cost method to use. It can be provided as a string.
            See `CostMethod`. It also indicates what is the correct unit of `association_threshold`.
            Default: EUCLIDEAN
        track_building (TrackBuilding): Tells the linker how to build the final tracks.
            Either from detections, or from filtered/smoothed positions computed by the
            Kalman filter. See `TrackBuilding`. It can be provided as a string.
            Default: FILTERED
        extract_flows_on_detections (bool): If True it extracts the optical flow at the detection location if possible.
            Otherwise it extract the flow from the curent estimate of the track position.
            Default: False
        always_measure_velocity (bool): Update velocity for all tracks even non-linked ones.
            If set to False, it implements KOFT-- from the paper. This is sub-optimal, you should keep it True.
            Default: True

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        association_threshold: float,
        detection_std: Union[float, torch.Tensor] = 3.0,
        flow_std: Union[float, torch.Tensor] = 1.0,
        process_std: Union[float, torch.Tensor] = 1.5,
        kalman_order: int = 1,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        cost: Union[str, Cost] = Cost.EUCLIDEAN,
        track_building: Union[str, TrackBuilding] = TrackBuilding.FILTERED,
        extract_flows_on_detections=False,
        always_measure_velocity=True,
    ):
        super().__init__(
            association_threshold,
            detection_std,
            process_std,
            kalman_order,
            n_valid,
            n_gap,
            association_method,
            cost,
            track_building,
        )

        self.flow_std = flow_std
        self.extract_flows_on_detections = extract_flows_on_detections
        self.always_measure_velocity = always_measure_velocity

        assert self.kalman_order >= 1, "With KOFT, the velocity is measured and thus should be modeled."

    flow_std: Union[float, torch.Tensor]
    extract_flows_on_detections: bool
    always_measure_velocity: bool


class KOFTLinker(KalmanLinker):
    """Kalman and Optical Flow Tracking [9]

    Motion is modeled with a Kalman filter of a specified order >= 1 (See `torch_kf.ckf`)
    Positions are measured through the detection process. A second update step is performed
    to measure the velocity of all tracks using optical flow.

    Matching is done to optimize the given cost.

    Note:
        This implementation requires torch-kf. (pip install torch-kf)

    See `KalmanLinker` for the other attributes.

    Attributes:
        specs (KOFTLinkerParameters): Parameters specifications of the algorithm.
            See `KOFTLinkerParameters`.
        last_detections (byotrack.Detections): The last detections used in update.
            Optionnaly used to extract flows at the detection positions and not the track state.
            Required for `motion_model`
        n_initial (int): Number of newly started tracks on the last update. Used to correclty
            initialized these tracks with the velocity update.

    """

    progress_bar_description = "KOFT linking"

    def __init__(
        self, specs: KOFTLinkerParameters, optflow: Optional[byotrack.OpticalFlow] = None, save_all=False
    ) -> None:
        super().__init__(specs, optflow, save_all)

        self.specs: KOFTLinkerParameters
        assert self.optflow is not None, "KOFT requires an optical flow algorithm"
        self.optflow: OnlineFlowExtractor

        self.last_detections = byotrack.Detections(data={"position": torch.empty((0, 2))})
        self.n_initial = 0

    def reset(self) -> None:
        super().reset()
        self.last_detections = byotrack.Detections(data={"position": torch.empty((0, 2))})
        self.n_initial = 0

    def motion_model(self) -> None:
        # Not initialized yet
        if self.active_states is None or self.kalman_filter is None:
            return

        if self.optflow.flow_map is None:
            raise RuntimeError("The motion model cannot be applied without a computed flow")

        dim = self.kalman_filter.measure_dim // 2

        # Second update using velocity just before the prediction time
        points = self.active_states.mean[:, :dim, 0]
        if self.specs.extract_flows_on_detections:
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] != -1:
                    points[i] = self.last_detections.position[track.detection_ids[-1]]

        if not self.specs.always_measure_velocity:  # KOFT--
            mask = torch.full((len(self.active_tracks),), True)
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] == -1:
                    mask[i] = False

            points = points[mask]

        velocities = torch.tensor(self.optflow.optflow.flow_at(self.optflow.flow_map, points.numpy()))[..., None]

        if not self.specs.always_measure_velocity:  # KOFT--
            self.active_states[mask] = self.kalman_filter.update(
                self.active_states[mask],
                velocities,
                measurement_matrix=self.kalman_filter.measurement_matrix[dim:],
                measurement_noise=self.kalman_filter.measurement_noise[dim:, dim:],
            )
        else:  # KOFT
            self.active_states = self.kalman_filter.update(
                self.active_states,
                velocities,
                measurement_matrix=self.kalman_filter.measurement_matrix[dim:],
                measurement_noise=self.kalman_filter.measurement_noise[dim:, dim:],
            )

        # Correct initial tracks: The velocity of initial tracks is initialize with computed flow
        # and the flow_std uncertainty
        if self.n_initial:
            self.active_states.mean[-self.n_initial :, dim : dim * 2] = velocities[-self.n_initial :]
            self.active_states.covariance[-self.n_initial :, dim : dim * 2, dim : dim * 2] = (
                self.kalman_filter.measurement_noise[dim:, dim:]
            )
        # Use the Kalman filter to predict the current states of each active tracks
        self.active_states = self.kalman_filter.predict(self.active_states)

        # Project states for association
        self.projections = self.kalman_filter.project(self.active_states)

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        dim = detections.position.shape[1]
        if self.active_states is None or self.kalman_filter is None:
            self.kalman_filter = torch_kf.ckf.constant_kalman_filter(
                self.specs.detection_std,
                self.specs.process_std,
                dim,
                self.specs.kalman_order,
                approximate=True,  # Approximate so that a flow precisely means the velocity modeled here
            )
            # Doubles the measurement space to measure velocity
            self.kalman_filter.measurement_matrix = torch.eye(dim * 2, self.kalman_filter.state_dim)
            self.kalman_filter.measurement_noise = torch.eye(dim * 2)
            self.kalman_filter.measurement_noise[:dim, :dim] *= self.specs.detection_std**2
            self.kalman_filter.measurement_noise[dim:, dim:] *= self.specs.flow_std**2

            self.active_states = torch_kf.GaussianState(
                torch.empty((0, self.kalman_filter.state_dim, 1)),
                torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim)),
            )
            self.projections = self.kalman_filter.project(self.active_states)

        if self.projections is None:
            raise RuntimeError("Projections should already be initialized.")

        if self.specs.cost == Cost.EUCLIDEAN:
            return torch.cdist(self.projections.mean[:, :dim, 0], detections.position), self.specs.association_threshold

        if self.specs.cost == Cost.EUCLIDEAN_SQ:
            return (
                torch.cdist(self.projections.mean[:, :dim, 0], detections.position).pow_(2),
                self.specs.association_threshold**2,
            )

        # Restrict projections onto positions
        # NOTE: Following KOFT official implem, we use (cov-1)[:2, :2] instead of (cov[:2, :2])-1 for the precision.
        projections = torch_kf.GaussianState(
            self.projections.mean[:, :dim],
            self.projections.covariance[:, :dim, :dim],
            self.projections.precision[:, :dim, :dim] if self.projections.precision is not None else None,
        )

        if self.specs.cost == Cost.MAHALANOBIS:
            return (
                projections[:, None].mahalanobis(detections.position[None, ..., None]),
                self.specs.association_threshold,
            )
        if self.specs.cost == Cost.MAHALANOBIS_SQ:
            return (
                projections[:, None].mahalanobis_squared(detections.position[None, ..., None]),
                self.specs.association_threshold**2,
            )

        # LIKELIHOOD: cost = -log likelihood
        cost = -projections[:, None].log_likelihood(detections.position[None, ..., None])
        return cost, -torch.log(torch.tensor(self.specs.association_threshold)).item()

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, links: torch.Tensor):
        if self.active_states is None or self.kalman_filter is None or self.projections is None:
            raise RuntimeError("The linker should already be initialized.")

        self.last_detections = detections  # Save detections (May be required)

        dim = detections.position.shape[1]

        # Update the state of associated tracks (unassociated tracks keep the predicted state)
        self.active_states[links[:, 0]] = self.kalman_filter.update(
            self.active_states[links[:, 0]],
            detections.position[links[:, 1]][..., None],
            torch_kf.GaussianState(
                self.projections.mean[links[:, 0], :dim],
                self.projections.covariance[links[:, 0], :dim, :dim],
                None,  # /!\ inv(cov[:2,:2]) != inv(cov)[:2, :2] =>
            ),
            self.kalman_filter.measurement_matrix[:dim],
            self.kalman_filter.measurement_noise[:dim, :dim],
        )

        # Update active track handlers
        active_mask = self.update_active_tracks(links)

        # Create new track handlers for unmatched detections
        unmatched = self.handle_extra_detections(detections, links)
        unmatched_measures = detections.position[unmatched]
        self.n_initial = unmatched_measures.shape[0]

        # Build the initial states for tracks:
        # We initialize the position using the detection position and the measurement std as covariance.
        # For velocity, we also will have a measurement at prediction time, it will be therefore initialized
        # correctly at this moment. If acceleration or jerk is modeled, we set them at 0 with a large covariance.
        # In SKT, having a too large uncertainty on non measured states will produce very uncertain projections
        # which are either unlinkable with likelihood or too linkable with mahalanobis.
        # In KOFT as velocity is measured before the prediction/projection step, the projection is never fully
        # uncertain, therefore we can initialize them with as much uncertainty as required.
        initial_state = torch_kf.GaussianState(
            torch.zeros(len(unmatched_measures), self.kalman_filter.state_dim, 1),
            torch.eye(self.kalman_filter.state_dim)[None].expand(
                len(unmatched_measures), self.kalman_filter.state_dim, self.kalman_filter.state_dim
            )
            * self.kalman_filter.process_noise.max()
            * 100,
        )
        initial_state.mean[:, :dim, 0] = unmatched_measures
        initial_state.covariance[:, :dim, :dim] = self.kalman_filter.measurement_noise[:dim, :dim]

        # Merge still active states with the initial ones of the created tracks
        self.active_states = torch_kf.GaussianState(
            torch.cat((self.active_states.mean[active_mask], initial_state.mean)),
            torch.cat((self.active_states.covariance[active_mask], initial_state.covariance)),
        )

        if self.specs.track_building == TrackBuilding.DETECTION:
            self.all_positions.append(
                torch.cat(
                    [
                        (
                            detections.position[track.detection_ids[-1]][None]
                            if track.detection_ids[-1] != -1
                            else torch.full((1, detections.position.shape[1]), torch.nan)
                        )
                        for track in self.active_tracks
                    ]
                )
            )
        else:
            self.all_positions.append(self.active_states.mean[:, :2, 0])

        if self.save_all or self.specs.track_building == TrackBuilding.SMOOTHED:
            self.all_states.append(self.active_states)



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import scipy.ndimage  # type: ignore


class OpticalFlow(ABC):
    """Optical flow algorithm (Abstract class, see implementations)

    It implements or wraps an existing method.
    In ByoTrack, frames are multidimensional floating arrays, they are
    the inputs of the optical flow algorithm. It outputs an optical flow map which
    is a multidimensional floating array where the first dimension is the flow for
    each spatial dimension.

    For instance, with two dimensional frames (H, W, C), the computed flow map has a shape
    (2, H, W), where the first (resp. second) element of the first dimension is the displacement
    along height (resp. width). The coordinate system of ByoTrack is using the pixel index (i, j) and not
    the pixel position (x, y) (which is often use in OpenCv)

    Additionally, the wrapper allows to downscale (with a blurring) the image to speed up computations.

    Usage:

    .. code-block:: python

        video: Video
        optflow: OpticalFlow

        # Preprocess the frames (reference and moving)
        ref = optflow.preprocess(video[0])
        mov = optflow.preprocess(video[1])

        # Compute the flow map from ref to mov
        flow_map = optflow.compute(ref, mov)

        # flow_map[:, i, j] gives the displacement (di, dj) of the pixel (i, j) of the reference image

        # You can warp mov into ref with warp
        mov_warp = optflow.warp(flow_map, video[1])

        # Or use the flow to move points


    Attributes:
        downscale (Union[float, np.ndarray]): Downscale factor. Can be provided for each spatial axis.
            Default: 2.0
        blur (Union[None, float, np.ndarray]): Std of the Gaussian blur applied before downscaling the images
            Can be provided for each spatial axis. If not provided, a default one is choosen given the downscale.

    """

    def __init__(self, downscale: Union[float, np.ndarray] = 2.0, blur: Union[None, float, np.ndarray] = None):
        self.downscale = downscale
        self.blur = blur

    def preprocess(self, frame: np.ndarray) -> np.ndarray:
        """Downscale a frame before running the optical flow algorithm

        It blurs then downscales the frame.

        Args:
            frame (np.ndarray): The frame to pre-preprocess
                Shape: (H, W, C), dtype: float64

        Returns:
            np.ndarray: The blurred and downscaled frame
                Shape: (H', W', C), dtype: float32

        """
        downscale = np.ones(frame.ndim, np.float32)
        downscale[:-1] = self.downscale  # Do not downscale channels

        blur = np.zeros(frame.ndim, np.float32)
        if self.blur is None:
            blur[:-1] = 2 * downscale[:-1] / 6
        else:
            blur[:-1] = self.blur

        # Blur
        frame = scipy.ndimage.gaussian_filter(frame.astype(np.float32), blur, mode="reflect")

        # Downscale (Linear interpolation)
        return scipy.ndimage.zoom(frame, 1 / downscale, order=1, mode="mirror", grid_mode=True)

    @abstractmethod
    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        """Compute the optical flow map from reference to the moving frame

        It computes the displacement of each pixel from the reference frame to be in the moving one.

        Args:
            reference (np.ndarray): Reference frame
                Shape: (H', W', C), dtype: float
            moving (np.ndarray): Moving frame
                Shape: (H', W', C), dtype: float

        Returns:
            np.ndarray: Optical flow map from reference to moving
                The pixel coordinates are stored as (i, j) (and not (x, y)), so is
                their displacement.
                Shape: (2, H', W'), dtype: float32
        """

    def flow_at(self, flow_map: np.ndarray, points: np.ndarray) -> np.ndarray:
        """Extract the flow/displacement at the given location

        The output displacement and the given location are expected to be at the frame scale.

        Args:
            flow_map (np.ndarray): Optical flow map (Displacement in i, j for each pixel)
                Shape: (2, H', W'), dtype: float32
            points (np.ndarray): Points where to extract the displacement (i, j) (Not downscaled)
                Shape: (N, 2), dtype: float

        Returns:
            np.ndarray: The displacement at points (i, j) (Not downscaled)
                Shape: (N, 2), dtype: float
        """
        if isinstance(self.downscale, (int, float)):
            downscale = np.array([self.downscale] * flow_map.shape[0])
        else:
            downscale = self.downscale

        flow = np.empty_like(points)

        for axis in range(flow_map.shape[0]):
            flow[:, axis] = scipy.ndimage.map_coordinates(flow_map[axis], points.T / downscale[axis])

        flow *= downscale

        return flow

    def transform(self, flow_map: np.ndarray, points: np.ndarray) -> np.ndarray:
        """Apply the flow to the given points

        The points are expected to be at the frame scale (not downscaled)

        Args:
            flow_map (np.ndarray): Optical flow map (Displacement in i, j for each pixel)
                Shape: (2, H', W'), dtype: float32
            points (np.ndarray): Points to transform (i, j) (Not downscaled)
                Shape: (N, 2), dtype: float

        Returns:
            np.ndarray: New positions of the points after the flow
                Shape: (N, 2), dtype: float
        """
        return points + self.flow_at(flow_map, points)

    def warp(self, flow_map: np.ndarray, moving: np.ndarray) -> np.ndarray:
        """Warps the moving image onto the reference using the flow map

        Note:
            We only implemented backward warping (It only allows to warp the moving image onto the reference one)
            Forward warping could be implemented in a future version if needed.

        Args:
            flow_map (np.ndarray): Optical flow map. Displacement (di, dj) for each pixel (i, j) of the
                reference frame.
                Shape: (2, H, W), dtype: float32
            moving (np.ndarray): The moving image to warp.
                Shape: (H, W, C]), dtype: float

        Returns:
            np.ndarray: Warped image
                Shape: (H, W, C), dtype: float

        """
        *dims, channels = moving.shape
        points = np.indices(dims, np.float32).reshape(2, -1).transpose()
        points = self.transform(flow_map, points).transpose().reshape(2, *dims)

        return np.concatenate(
            [
                scipy.ndimage.map_coordinates(moving[..., channel], points, mode="nearest", order=1)[..., None]
                for channel in range(channels)
            ],
            axis=-1,
        )


class DummyOpticalFlow(OpticalFlow):
    """Dummy optical flow which predict no displacement for each pixel"""

    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        return np.zeros((reference.ndim - 1, *reference.shape[:-1]))



import dataclasses
import enum
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch_kf
import torch_kf.ckf

import byotrack

from .nearest_neighbor import AssociationMethod, FrameByFrameLinker, FrameByFrameLinkerParameters


class Cost(enum.Enum):
    """The cost to use for association

    * LIKELIHOOD
        Build the cost matrix to maximize the likelihood of the association.
        The cost is defined as: the negative log likelihood that the detections j is from track i.
        C[i, j] = - log P(det_j | track_i)
        The cost limit is expected to be the smallest probability to accept.
    * MAHALANOBIS
        Mahalanobis distance, using the track uncertainty to correct the euclidean distance.
        The cost limit is expected to be the largest Mahalanobis distance to accept.
    * MAHALANOBIS_SQ
        Squared Mahalanobis distance.
        The cost limit is expected to be the largest Mahalanobis distance to accept. (not squared)
    * EUCLIDEAN
        Euclidean distance.
        The cost limit is expected to be the largest Euclidean distance to accept.
    * EUCLIDEAN_SQ
        Squared euclidean distance.
        The cost limit is expected to be the largest Euclidean distance to accept. (not squared)

    """

    LIKELIHOOD = "likelihood"
    MAHALANOBIS = "mahalanobis"
    MAHALANOBIS_SQ = "mahalanobis_sq"
    EUCLIDEAN = "euclidean"
    EUCLIDEAN_SQ = "euclidean_sq"


class TrackBuilding(enum.Enum):
    """How to build the final tracks

    * DETECTION
        Build tracks from detections without filtering nor filling gaps
    * FILTERED
        Build tracks from the Kalman filter outputs
    * SMOOTHED
        Add a backward run of the Kalman filter (RTS) to smooth tracks position

    """

    DETECTION = "detection"
    FILTERED = "filtered"
    SMOOTHED = "smoothed"


@dataclasses.dataclass
class KalmanLinkerParameters(FrameByFrameLinkerParameters):
    """Parameters of KalmanLinker

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
        detection_std (Union[float, torch.Tensor]): Expected measurement noise on the detection process.
            The detection process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension). See `torch_kf.ckf.constant_kalman_filter`.
            Default: 3.0 pixels
        process_std (Union[float, torch.Tensor]): Expect process noise. See `torch_kf.ckf.constant_kalman_filter`, the
            process is modeled as constant order-th derivative motion. This quantify how much the supposely "constant"
            order-th derivative can change between two consecutive frames. A common rule of thumb is to use
            3 * process_std ~= max_t(| x^(order)(t) - x^(order)(t+1)|). It can be provided for each dimension).
            Default: 1.5 pixels / frame^order
        kalman_order (int): Order of the Kalman filter to use.
            0 for brownian motions, 1 for directed brownian motion, 2 for accelerated brownian motions, etc...
            Default: 1
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
            Default: 3
        n_gap (int): Number of frames with no association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
        cost_method (CostMethod): The cost method to use. It can be provided as a string.
            See `CostMethod`. It also indicates what is the correct unit of `association_threshold`.
            Default: EUCLIDEAN
        track_building (TrackBuilding): Tells the linker how to build the final tracks.
            Either from detections, or from filtered/smoothed positions computed by the
            Kalman filter. See `TrackBuilding`. It can be provided as a string.
            Default: FILTERED

    """

    def __init__(
        self,
        association_threshold: float,
        detection_std: Union[float, torch.Tensor] = 3.0,
        process_std: Union[float, torch.Tensor] = 1.5,
        kalman_order: int = 1,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        cost: Union[str, Cost] = Cost.EUCLIDEAN,
        track_building: Union[str, TrackBuilding] = TrackBuilding.FILTERED,
    ):
        super().__init__(association_threshold, n_valid, n_gap, association_method)

        self.detection_std = detection_std
        self.process_std = process_std
        self.kalman_order = kalman_order

        self.cost = cost if isinstance(cost, Cost) else Cost[cost.upper()]
        self.track_building = (
            track_building if isinstance(track_building, TrackBuilding) else TrackBuilding[track_building.upper()]
        )

    detection_std: Union[float, torch.Tensor]
    process_std: Union[float, torch.Tensor]
    kalman_order: int
    cost: Cost
    track_building: TrackBuilding


class KalmanLinker(FrameByFrameLinker):
    """Frame by frame linker using Kalman filters

    Motion is modeled with a Kalman filter of a specified order (See `torch_kf.ckf`)
    Matching is done to optimize the given cost. If optical flow is provided, it is used
    online to warp the predicted state positions of the kalman filter. This will work, but it
    is sub-optimal: consider using `KOFTLinker` that exploits in a finer way optical flow
    inside Kalman filters.

    This is an implementation of Simple Kalman Tracking (SKT) from KOFT [9].

    Note:
        This implementation requires torch-kf. (pip install torch-kf)

    See `FrameByFrameLinker` for the other attributes.

    Attributes:
        specs (KalmanLinkerParameters): Parameters specifications of the algorithm.
            See `KalmanLinkerParameters`.
        kalman_filter (Optional[torch_kf.KalmanFilter]): The Kalman filter. (Build once the tracking starts
            allowing to adapt the dimension on the fly)
        active_states (Optional[torch_kf.GaussianState]): The Kalman filter estimation for each track.
            Shape: mean=(N, D * (order + 1), 1), covariance=(N, D * (order + 1), dim * (order + 1))
            dtype: float32
        projections (Optional[torch_kf.GaussianState]): The Kalman filter projection for each track.
            Shape: mean=(N, D, 1), covariance=(N, D, D), precision=(N, D, D)
            dtype: float32
        all_states (List[torch_kf.GaussianState]): The Kalman filter estimation for each track at each seen
            frame. States are only registered when save_all=True or if you build tracks from RTS smoothing.
            Shape: mean=(N, D * (order + 1), 1), covariance=(N, D * (order + 1), dim * (order + 1))
            dtype: float32

    """

    progress_bar_description = "Kalman filter linking"

    def __init__(
        self, specs: KalmanLinkerParameters, optflow: Optional[byotrack.OpticalFlow] = None, save_all=False
    ) -> None:
        super().__init__(specs, optflow, save_all)

        self.specs: KalmanLinkerParameters
        self.kalman_filter: Optional[torch_kf.KalmanFilter] = None
        self.active_states: Optional[torch_kf.GaussianState] = None
        self.projections: Optional[torch_kf.GaussianState] = None

        self.all_states: List[torch_kf.GaussianState] = []

    def reset(self) -> None:
        super().reset()

        self.kalman_filter = None
        self.active_states = None
        self.projections = None

        self.all_states = []

    def collect(self) -> List[byotrack.Track]:
        if self.specs.track_building == TrackBuilding.SMOOTHED:
            # TODO: Implement RTS + Add a tracker
            raise NotImplementedError("Smoothing is not implemented yet")
        return super().collect()

    def motion_model(self) -> None:
        # Not initialized yet
        if self.active_states is None or self.kalman_filter is None:
            return

        # Use the Kalman filter to predict the current states of each active tracks
        self.active_states = self.kalman_filter.predict(self.active_states)

        # Add optical flow motion to the position
        if self.optflow and self.optflow.flow_map is not None:
            positions = self.active_states.mean[:, : self.kalman_filter.measure_dim, 0]
            positions[:] = torch.tensor(self.optflow.optflow.transform(self.optflow.flow_map, positions.numpy()))

        # Project states for association
        self.projections = self.kalman_filter.project(self.active_states)

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        if self.active_states is None or self.kalman_filter is None:
            self.kalman_filter = torch_kf.ckf.constant_kalman_filter(
                self.specs.detection_std,
                self.specs.process_std,
                detections.position.shape[1],
                self.specs.kalman_order,
            )
            self.active_states = torch_kf.GaussianState(
                torch.empty((0, self.kalman_filter.state_dim, 1)),
                torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim)),
            )
            self.projections = self.kalman_filter.project(self.active_states)

        if self.projections is None:
            raise RuntimeError("Projections should already be initialized.")

        if self.specs.cost == Cost.EUCLIDEAN:
            return torch.cdist(self.projections.mean[..., 0], detections.position), self.specs.association_threshold

        if self.specs.cost == Cost.EUCLIDEAN_SQ:
            return (
                torch.cdist(self.projections.mean[..., 0], detections.position).pow_(2),
                self.specs.association_threshold**2,
            )
        if self.specs.cost == Cost.MAHALANOBIS:
            return (
                self.projections[:, None].mahalanobis(detections.position[None, ..., None]),
                self.specs.association_threshold,
            )
        if self.specs.cost == Cost.MAHALANOBIS_SQ:
            return (
                self.projections[:, None].mahalanobis_squared(detections.position[None, ..., None]),
                self.specs.association_threshold**2,
            )

        # LIKELIHOOD: cost = -log likelihood
        cost = -self.projections[:, None].log_likelihood(detections.position[None, ..., None])
        return cost, -torch.log(torch.tensor(self.specs.association_threshold)).item()

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, links: torch.Tensor):
        if self.active_states is None or self.kalman_filter is None or self.projections is None:
            raise RuntimeError("The linker should already be initialized.")

        # Update the state of associated tracks (unassociated tracks keep the predicted state)
        self.active_states[links[:, 0]] = self.kalman_filter.update(
            self.active_states[links[:, 0]], detections.position[links[:, 1]][..., None], self.projections[links[:, 0]]
        )

        # Update active track handlers
        active_mask = self.update_active_tracks(links)

        # Create new track handlers for unmatched detections
        unmatched = self.handle_extra_detections(detections, links)
        unmatched_measures = detections.position[unmatched]

        # Build the initial states for tracks:
        # We initialize the position using the detection position and the measurement std as covariance.
        # For velocity/acceleration, let's predict them at 0 but with 10 times the process_noise (without correlations)
        # But setting a too large uncertainty on the initial velocity produces a uncertain initial projection
        # that is hard to linked with likelihood/mahalanobis distance. (10 times the process noise seems to be
        # a good tradeoff)
        initial_state = torch_kf.GaussianState(
            torch.zeros(len(unmatched_measures), self.kalman_filter.state_dim, 1),
            self.kalman_filter.process_noise[None].expand(
                len(unmatched_measures), self.kalman_filter.state_dim, self.kalman_filter.state_dim
            )
            * torch.eye(self.kalman_filter.state_dim)
            * 10,
        )
        initial_state.mean[:, : unmatched_measures.shape[1], 0] = unmatched_measures
        initial_state.covariance[:, : unmatched_measures.shape[1], : unmatched_measures.shape[1]] = (
            self.kalman_filter.measurement_noise
        )

        # Merge still active states with the initial ones of the created tracks
        self.active_states = torch_kf.GaussianState(
            torch.cat((self.active_states.mean[active_mask], initial_state.mean)),
            torch.cat((self.active_states.covariance[active_mask], initial_state.covariance)),
        )

        if self.specs.track_building == TrackBuilding.DETECTION:
            self.all_positions.append(
                torch.cat(
                    [
                        (
                            detections.position[track.detection_ids[-1]][None]
                            if track.detection_ids[-1] != -1
                            else torch.full((1, detections.position.shape[1]), torch.nan)
                        )
                        for track in self.active_tracks
                    ]
                )
            )
        else:
            self.all_positions.append(self.active_states.mean[:, :2, 0])

        if self.save_all or self.specs.track_building == TrackBuilding.SMOOTHED:
            self.all_states.append(self.active_states)



from typing import Collection, List, Sequence, Union

import numpy as np
import torch

import byotrack

from . import propagation


class ForwardBackwardInterpolater(byotrack.Refiner):  # pylint: disable=too-few-public-methods
    """Interpolate tracks to fill out NaN values by averaging a forward estimation and a backward estimation.

    We propose three modes for position propagation: constant, tps and flow. Constant estimation propagates the
    last known position, tps interpolates elastically the motion of the active tracks to fill missing ones and
    flow uses optical flow to propagate the tracks.

    Forward and backward propagation are smoothly merged together in a weighted average. See `propagation` module

    Attributes:
        method (str | DirectedPropagate): Method to use for position propagation. Either "constant", "tps"
            or "flow" (see `constant_directed_propagate`, `tps_directed_propagate` or `optical_flow_directed_propagate`)
            or a user defined Callable following the `DirectedPropagate` protocol.
            The method will be called with the tracks_matrix, video, a boolean direction and additional kwargs.
        full (bool): If True, it will extrapolate positions for all the frames of the video
            If False, it will only interpolate NaN values inside each track domain, without extrapolation
            Default: False
        kwargs (Dict[str, Any]): Additional parameters given to `method`.
            For instance for "tps", you can set `alpha` (float) which is the regularization parameter of the
            interpolation (see `tps_propation` module and `torch_tps` librarie). We advise to use alpha > 5.0
            (alpha = 10 by default). For "flow" propagation, you should provide `optflow` (byotrack.OpticalFlow)
            to correctly estimate optical flows.

    """

    def __init__(self, method: Union[str, propagation.DirectedPropagate] = "constant", full=False, **kwargs) -> None:
        super().__init__()
        self.method = method
        self.full = full
        self.kwargs = kwargs

    def run(
        self, video: Union[Sequence[np.ndarray], np.ndarray], tracks: Collection[byotrack.Track]
    ) -> List[byotrack.Track]:
        if not tracks:
            return []

        tracks_matrix = byotrack.Track.tensorize(tracks, frame_range=(0, len(video)) if self.full else None)
        if not self.full:
            start = min(track.start for track in tracks)
            end = max(track.start + len(track) for track in tracks)
            video = video[start:end]  # Clip video temporally so that it matches with the tracks_matrix
        propagation_matrix = propagation.forward_backward_propagation(tracks_matrix, video, self.method, **self.kwargs)

        new_tracks = []
        for i, track in enumerate(tracks):
            if self.full:
                start = 0
                points = propagation_matrix[:, i]
                detection_ids = torch.full((propagation_matrix.shape[0],), -1, dtype=torch.int32)
                detection_ids[track.start : track.start + len(track)] = track.detection_ids
            else:
                start = track.start
                points = propagation_matrix[start : start + len(track), i]
                detection_ids = track.detection_ids.clone()
            new_tracks.append(byotrack.Track(start, points, track.identifier, detection_ids))

        return new_tracks



from __future__ import annotations

from abc import ABC, abstractmethod

from typing import Any, Dict, Set


class ParameterRange(ABC):
    """Define a range of validity for parameters"""

    @abstractmethod
    def is_valid(self, parameter: object) -> bool:
        """Check that the parameter's value is in the validity range"""

    @abstractmethod
    def to_str(self) -> str:
        """Returns a string describing the valid range"""


class ParameterBound(ParameterRange):
    """Bounded parameters between low and high values"""

    def __init__(self, low: float, high: float) -> None:
        super().__init__()
        self.low = low
        self.high = high

    def is_valid(self, parameter: object) -> bool:
        if isinstance(parameter, (float, int)):
            return self.low <= parameter <= self.high
        return False

    def to_str(self) -> str:
        return f"[{self.low}, {self.high}]"


class ParameterEnum(ParameterRange):
    """Parameters that can have only a set of predefined value"""

    def __init__(self, values: Set) -> None:
        super().__init__()
        self.values = values

    def is_valid(self, parameter: object) -> bool:
        return parameter in self.values

    def to_str(self) -> str:
        return str(self.values)


class ParametrizedObjectMixin:  # pylint: disable=too-few-public-methods
    """Objects with hyper parameters

    The child classes have to define the `parameters` dict that describes all their parameters
    with a ParameterRange. When the instance tries to set a parameter value, this range is used
    to validate the value.

    (See an implementation with SpotDetector)

    Attributes:
        parametrized (Dict[str, ParametrizedObjectMixin]): Registered child ParametrizedObjectMixin
            ParametrizedObjectMixin attributes are automatically registered.
        parameters (Dict[str, ParameterRange]): Defined range for parameters

    """

    parametrized: Dict[str, ParametrizedObjectMixin] = {}
    parameters: Dict[str, ParameterRange] = {}

    def __setattr__(self, __name: str, __value: Any) -> None:
        if isinstance(__value, ParametrizedObjectMixin):
            self.parametrized[__name] = __value

        parameter_range = self.parameters.get(__name)
        if parameter_range is None or parameter_range.is_valid(__value):
            return super().__setattr__(__name, __value)

        raise ValueError(f"Invalid value for parameter {__name}: {__value} not in {parameter_range.to_str()}")



from typing import cast, List, Optional, Union

import cv2  # type: ignore
import numpy as np
import torch

import byotrack
from byotrack.api.detector.detections import relabel_consecutive
from byotrack.api.parameters import ParameterBound, ParameterEnum


# Set to true to follow closely the ICY implementation (slower and less precise)
FOLLOW_ICY = False


class B3SplineUWT(torch.nn.Module):
    """Undecimated Wavelet Transform with B3Spline for 2D images

    Also called A trous Wavelet Transform.

    Let J be the level of the UWD.
    Let c_0 be the original image. It computes successively dilated convolution of it.

    c_j = h_j * c_{j-1}, 1 \\le j \\le J
    w_j = c_{j-1} - c_{j}, 1 \\le j \\le J

    Where h_j is the filter at scale j which is the original filter h_1
    dilated with 2^{j - 1} 0 between each coefficient and h_1 = [1/16, 1/4, 3/8, 1/4, 1/16].

    The parameters returned are [w_1, w_2, ..., w_J, c_J].
    The original image is easily reconstructed from the parameters with c_0 = c_J + \\sum_j w_j

    Boundaries issues are handled using mirror padding.

    """

    weights = torch.tensor([1 / 16, 1 / 4, 3 / 8, 1 / 4, 1 / 16])

    def __init__(self, level: int = 3):
        """Constructor

        Args:
            level (int): Level of the Wavelet Transform. (J)

        """
        super().__init__()
        self.layers = torch.nn.ModuleList(
            [
                torch.nn.Conv2d(1, 1, 5, dilation=2**i, padding=2 ** (i + 1), bias=False, padding_mode="reflect")
                for i in range(level)
            ]
        )

        weights_2d = self.weights[None, :] * self.weights[:, None]  # 2D equivalent for the B3Spline

        for layer in self.layers:
            layer.weight.requires_grad_(False)
            layer.weight[0, 0] = weights_2d

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the parameters of the UWD

        Args:
            x (torch.Tensor): Input images without channel dimension
                Shape: (B, H, W)

        Returns:
            torch.Tensor: Parameters [w_1, ..., w_J, c_J]
                Shape: (B, J + 1, H, W)

        """
        assert len(x.shape) == 3

        x = x[:, None, :, :]  # Shape: B x 1 x H x W

        outputs = torch.zeros((x.shape[0], len(self.layers) + 1, x.shape[-2], x.shape[-1]), device=x.device)

        for j, layer in enumerate(self.layers):
            y = layer(x)
            outputs[:, j] = (x - y)[:, 0]
            x = y

        outputs[:, -1] = x[:, 0]

        return outputs


class B3SplineUWTApprox(torch.nn.Module):
    """Approximation of Undecimated Wavelet Transform with B3Spline for 2D images

    Split the 2D convolution in two 1D convolution first alongside the rows then the columns.

    After some analysis this implementation is slower than
    the first one... even though it should be faster.

    """

    weights = torch.tensor([1 / 16, 1 / 4, 3 / 8, 1 / 4, 1 / 16])

    def __init__(self, level=3):
        """Constructor

        Args:
            level (int): Level of the Wavelet Transform. (J)

        """
        super().__init__()
        self.layers = torch.nn.ModuleList(
            [
                torch.nn.Conv1d(1, 1, 5, dilation=2**i, padding=2 ** (i + 1), bias=False, padding_mode="reflect")
                for i in range(level)
            ]
        )

        for layer in self.layers:
            layer.weight.requires_grad_(False)
            layer.weight[0, 0] = self.weights

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the parameters of the UWD

        XXX: Batchsize != 1 is not supported yet

        Args:
            x (torch.Tensor): Input images without channel dimension
                Shape: B x H x W

        Returns:
            torch.Tensor: Parameters [w_1, ..., w_J, c_J]
                Shape: B x J + 1 x H x W

        """
        assert len(x.shape) == 3
        assert x.shape[0] == 1, "Batch size > 1 is not supported"

        x = x[0, :, None, :]  # H x 1 x W
        outputs = torch.zeros((len(self.layers) + 1, x.shape[-3], x.shape[-1]), device=x.device)

        for j, layer in enumerate(self.layers):
            y = layer(x).permute((2, 1, 0))  # W x 1 x H
            y = layer(y).permute((2, 1, 0))
            outputs[j] = (x - y)[:, 0]
            x = y

        outputs[-1] = x[:, 0]

        return outputs[None, ...]


class WaveletDetector(byotrack.BatchDetector):
    """Detection of bright spots using B3SplineUWT

    Following paper from Olivo-Marin, J.C. Extraction of spots in biological images using
    multiscale products. Pattern Recognit. 35, 1989-1996

    The algorithm is in 4 steps:

    1. UWT decomposition
    2. Scale selection
    3. Noise filtering
    4. Connected components extraction

    The multi scales behavior (choosing multiple scales) was implemented but we decided to drop it.
    It adds complexity without real advantages from our experience.

    The same algorithm is implemented in Icy Software (SpotDetector). The main differences are:

    * 2d wavelets (rather than 2 times one dimensional wavelets). It was designed to improve computations,
      but with torch no gain in time is observed. (Can be switch with FOLLOW_ICY)
    * Thresholding -> We follow the original paper.

    Attributes:
        scale (int): Scale of the wavelet coefficients used. With small scales, the detector focus on
            smaller objects.
        k (float): Noise threshold. Following the paper, the wavelet coefficients
            are filtered if coef \\le k \\sigma. (The higher the less spots you retrieve)
        min_area (float): Filter resulting spots that are too small (less than min_area pixels)
        device (torch.device): Device on which run the B3SplineUWT
            Default to cpu
        b3swt (B3SplineUWT): Undecimated wavelet transform
        **kwargs: Additional arguments for `BatchDetector` (batch_size, add_true_frames)

    Warning: The connected components used (opencv) yields segfault with too many components...

    """

    progress_bar_description = "Detections (Wavelet)"

    parameters = {
        "scale": ParameterEnum({0, 1, 2, 3, 4}),
        "k": ParameterBound(1.0, 10.0),
        "min_area": ParameterBound(0.0, 30.0),
    }

    def __init__(self, scale=2, k=3.0, min_area=10.0, device: Optional[torch.device] = None, **kwargs):
        super().__init__(**kwargs)
        self.scale = scale
        self.k = k
        self.min_area = min_area
        self.device = device if device else torch.device("cpu")
        self.b3swt: Union[B3SplineUWT, B3SplineUWTApprox]
        if FOLLOW_ICY:
            self.b3swt = B3SplineUWTApprox(scale + 1).to(self.device)
        else:
            self.b3swt = B3SplineUWT(scale + 1).to(self.device)

    def detect(self, batch: np.ndarray) -> List[byotrack.Detections]:
        assert (
            batch.shape[-1] == 1
        ), "This detector does not support multi channel images. Please aggregate the channels first"

        # Extract wavelet coefficient at the given scale
        inputs = torch.tensor(batch, dtype=torch.float32)[..., 0].to(self.device)
        coefficients: torch.Tensor = self.b3swt(inputs)[:, self.scale]

        # Thresholding. Shape: (B, H, W)
        coefficients[coefficients < self.compute_threshold(coefficients)] = 0
        coefficients[coefficients > 0] = 255

        # Instance segmentation by connected components
        # (Not found a quick and efficient torch implem with 4-way connectivity)
        detections_list = []
        mask: np.ndarray
        for i, mask in enumerate(coefficients.cpu().to(torch.uint8).numpy()):  # Unefficient but fine.
            _, segmentation, stats, _ = cv2.connectedComponentsWithStats(
                mask, np.zeros(mask.shape, dtype=np.uint16), connectivity=4, ltype=cv2.CV_16U  # type: ignore
            )

            # Fix typing issues of open-cv
            stats = cast(np.ndarray, stats)
            segmentation = cast(np.ndarray, segmentation)

            # Delete too small detections
            area = stats[:, 4]
            segmentation[(area < self.min_area)[segmentation]] = 0

            detections_list.append(
                byotrack.Detections(
                    {
                        "segmentation": relabel_consecutive(torch.tensor(segmentation.astype(np.int32))),
                        # "confidence": put areas instead of deleting them ?
                    },
                    frame_id=i,
                )
            )

        return detections_list

    def compute_threshold(self, coefficients: torch.Tensor) -> torch.Tensor:
        """Compute threshold for the UWT coefficients

        Note: One could use MAD approx of \\sigma rather than std (or Icy implem) but it's almost equivalent
        (And k is a parameter to tune so it truly is)

        Args:
            coefficients (torch.Tensor): Coefficients of the UWT
                Shape: (..., H, W)

        Returns:
            torch.Tensor: Threshold for each scale
                Shape: (..., 1, 1)

        """
        return coefficients.std(dim=(-1, -2), keepdim=True) * self.k
